#!/usr/bin/python3
"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""

import json
from pkscreener.classes.PKTask import PKTask
from PKDevTools.classes.SuppressOutput import SuppressOutput
from pkscreener.classes.PKScheduler import PKScheduler
from PKDevTools.classes.log import default_logger

class PKDataService():
    def getSymbolsAndSectorInfo(self,configManager,stockCodes=[]):
        from PKNSETools.PKCompanyGeneral import download, initialize
        stockDictList = []
        tasksList = []
        for symbol in stockCodes:
            fn_args = (symbol)
            task = PKTask(f"DataDownload-{symbol}",long_running_fn=download,long_running_fn_args=fn_args)
            task.userData = symbol
            tasksList.append(task)
        
        processedStocks = []
        if len(tasksList) > 0:
            # Suppress any multiprocessing errors/warnings
            with SuppressOutput(suppress_stderr=True, suppress_stdout=True):
                initialize() # Let's get the cookies set-up right
                PKScheduler.scheduleTasks(tasksList=tasksList,
                                        label=f"Downloading latest symbol/sector info. (Total={len(stockCodes)} records in {len(tasksList)} batches){'Be Patient!' if len(stockCodes)> 2000 else ''}",
                                        timeout=(5+2.5*configManager.longTimeout*4), # 5 seconds additional time for getting multiprocessing ready
                                        minAcceptableCompletionPercentage=100,
                                        submitTaskAsArgs=True,
                                        showProgressBars=True)
            for task in tasksList:
                if task.result is not None:
                    taskResult = json.loads(task.result)
                    if taskResult is not None and isinstance(taskResult,dict) and "info" in taskResult.keys():
                        stockDictList.append(taskResult.get("info"))
                        processedStocks.append(task.userData)
        leftOutStocks = list(set(stockCodes)-set(processedStocks))
        default_logger().debug(f"Attempted fresh download of {len(stockCodes)} stocks and downloaded {len(processedStocks)} stocks. {len(leftOutStocks)} stocks remaining.")
        return stockDictList, leftOutStocks



"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import warnings
from time import sleep

import numpy as np

warnings.simplefilter("ignore", DeprecationWarning)
warnings.simplefilter("ignore", FutureWarning)
import pandas as pd
from PKDevTools.classes.ColorText import colorText
from PKDevTools.classes.OutputControls import OutputControls

from pkscreener import Imports

if Imports["talib"]:
    try:
        import talib
    except Exception as e: # pragma: no cover
        issueLink = "https://github.com/pkjmesra/PKScreener"
        issueLink = f"\x1b[97m\x1b]8;;{issueLink}\x1b\\{issueLink}\x1b]8;;\x1b\\\x1b[0m"
        taLink = "https://github.com/ta-lib/ta-lib-python"
        taLink = f"\x1b[97m\x1b]8;;{taLink}\x1b\\{taLink}\x1b]8;;\x1b\\\x1b[0m"
        OutputControls().printOutput(
                colorText.FAIL
                + f"  [+] 'TA-Lib' library is not installed. For best results, please install 'TA-Lib'! You may wish to follow instructions from\n  [+] {issueLink}"
                + colorText.END
            )
        try:
            import pandas_ta_classic as talib
            OutputControls().printOutput(
                colorText.FAIL
                + f"  [+] TA-Lib is not installed. Falling back on pandas_ta_classic.\n  [+] For full coverage(candle patterns), you may wish to follow instructions from\n  [+] {taLink}"
                + colorText.END
            )
        except: # pragma: no cover
            OutputControls().printOutput(
                colorText.FAIL
                + f"  [+] pandas_ta_classic is not installed. Falling back on pandas_ta_classic also failed.\n  [+] For full coverage(candle patterns), you may wish to follow instructions from\n  [+] {taLink}"
                + colorText.END
            )
            pass
        pass
else:
    try:
        import pandas_ta_classic as talib
        OutputControls().printOutput(
            colorText.FAIL
            + "  [+] TA-Lib is not installed. Falling back on pandas_ta_classic.\n  [+] For full coverage(candle patterns), you may wish to follow instructions from\n  [+] https://github.com/ta-lib/ta-lib-python"
            + colorText.END
        )
        sleep(3)
    except Exception:  # pragma: no cover
        # default_logger().debug(e, exc_info=True)
        import talib


class pktalib:

    @classmethod
    def align_series(*series_list, fill_value=0):
        """
        Aligns multiple Pandas Series to have the same index and returns them.
        Missing values are filled with the given fill_value.
        """
        if all(isinstance(s, pd.Series) for s in series_list):
            aligned_series = pd.concat(series_list, axis=1).fillna(fill_value)
            return [aligned_series[col] for col in aligned_series.columns]
        return series_list

    @classmethod
    def AVWAP(self,df,anchored_date:pd.Timestamp):
        # anchored_date = pd.to_datetime('2022-01-30')
        # Choosing a meaningful anchor point is an essential part of using 
        # Anchored VWAP effectively. Traders may choose to anchor VWAP to 
        # a significant event that is likely to impact the stock’s price 
        # movement, such as an earnings announcement, product launch, or 
        # other news events. By anchoring VWAP to such events, traders 
        # can get a more meaningful reference point that reflects the 
        # sentiment of the market around that time. This can help traders 
        # identify potential areas of support and resistance more accurately 
        # and make better trading decisions.
        with pd.option_context('mode.chained_assignment', None):
            df["VWAP_D"] = pktalib.VWAP(high=df["high"],low=df["low"],close=df["close"],volume=df["volume"],anchor="D")
            # If we create a column 'typical_price', it should be identical with 'VWAP_D'
            df['typical_price'] = (df["high"] + df["low"] + df["close"])/3
            tpp_d = ((df["high"] + df["low"] + df["close"])*df["volume"])/3
            df['anchored_VWAP'] = tpp_d.where(df.index >= anchored_date).groupby(
                df.index >= anchored_date).cumsum()/df["volume"].where(
                    df.index >= anchored_date).groupby(
                        df.index >= anchored_date).cumsum()
        return df['anchored_VWAP']

    @classmethod
    def BBANDS(self, close, timeperiod,std=2, mamode=0):
        try:
            return talib.bbands(close, timeperiod)
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.BBANDS(close, timeperiod, std, std, mamode)
        
    @classmethod
    def EMA(self, close, timeperiod):
        try:
            return talib.ema(close, timeperiod)
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.EMA(close, timeperiod)

    @classmethod
    def VWAP(self, high, low, close, volume,anchor=None):
        try:
            import pandas_ta_classic as talib
            # Aligning the series
            # high,low,close = pktalib.align_series(high, low, close, fill_value=0)
            return talib.vwap(high, low, close, volume,anchor=anchor)
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return None
        
    @classmethod
    def KeltnersChannel(self, high, low, close, timeperiod=20):
        try:
            low_kel = None
            upp_kel = None
            # Aligning the series
            # high,low,close = pktalib.align_series(high, low, close, fill_value=0)
            tr = pktalib.TRUERANGE(high, low, close)
            atr = pktalib.ATR(high, low, close, timeperiod=timeperiod)
            sma = pktalib.SMA(close=close, timeperiod=timeperiod)
            low_kel = sma - atr * 1.5
            upp_kel = sma + atr * 1.5
            return low_kel, upp_kel
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return low_kel, upp_kel
        
    @classmethod
    def SMA(self, close, timeperiod):
        try:
            return talib.sma(close, timeperiod)
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.SMA(close, timeperiod)

    @classmethod
    def WMA(self, close, timeperiod):
        try:
            return talib.wma(close, timeperiod)
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.WMA(close, timeperiod)
        
    @classmethod
    def ATR(self, high, low, close, timeperiod=14):
        try:
            # Aligning the series
            # high,low,close = pktalib.align_series(high, low, close, fill_value=0)
            return talib.atr(high, low, close, length= timeperiod)
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.ATR(high, low, close, timeperiod=timeperiod)
        
    @classmethod
    def TRUERANGE(self, high, low, close):
        try:
            # Aligning the series
            # high,low,close = pktalib.align_series(high, low, close, fill_value=0)
            return talib.true_range(high, low, close)
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.TRANGE(high, low, close)

    @classmethod
    def MA(self, close, timeperiod):
        try:
            return talib.ma(close, timeperiod)
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.MA(close, timeperiod)

    @classmethod
    def TriMA(self, close,length=10):
        try:
            import pandas_ta_classic as talib
            return talib.trima(close=close, length=length)
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return None

    @classmethod
    def MACD(self, close, fast, slow, signal):
        try:
            # import pandas_ta_classic as talib
            return talib.macd(close, fast, slow, signal, talib=Imports["talib"])
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.MACD(close, fast, slow, signal)

    @classmethod
    def MFI(self, high, low, close,volume, timeperiod=14):
        try:
            # Aligning the series
            # high,low,close,volume = pktalib.align_series(high, low, close,volume, fill_value=0)
            return talib.mfi(high, low, close,volume, length= timeperiod)
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.MFI(high, low, close,volume, timeperiod=timeperiod)

    @classmethod
    def RSI(self, close, timeperiod):
        try:
            return talib.rsi(close, timeperiod)
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.RSI(close, timeperiod)

    @classmethod
    def CCI(self, high, low, close, timeperiod):
        try:
            # Aligning the series
            # high,low,close = pktalib.align_series(high, low, close, fill_value=0)
            return talib.cci(high, low, close, timeperiod)
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CCI(high, low, close, timeperiod)

    @classmethod
    def Aroon(self, high, low, timeperiod):
        try:
            # Aligning the series
            # high,low = pktalib.align_series(high, low, fill_value=0)
            return talib.aroon(high, low, timeperiod)
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            aroon_down, aroon_up = talib.AROON(high, low, timeperiod)
            aroon_up.name = f"AROONU_{timeperiod}"
            aroon_down.name = f"AROOND_{timeperiod}"
            data = {
                aroon_down.name: aroon_down,
                aroon_up.name: aroon_up,
            }
            return pd.DataFrame(data)

    @classmethod
    def STOCHF(self, high, low, close, fastk_period, fastd_period, fastd_matype):
        # Aligning the series
        # high,low,close = pktalib.align_series(high, low, close, fill_value=0)
        fastk, fastd = talib.STOCHF(high,
                            low,
                            close,
                            fastk_period, 
                            fastd_period,
                            fastd_matype)
        return fastk, fastd
    
    @classmethod
    def STOCHRSI(self, close, timeperiod, fastk_period, fastd_period, fastd_matype):
        try:
            _name = "STOCHRSI"
            _props = f"_{timeperiod}_{timeperiod}_{fastk_period}_{fastd_period}"
            stochrsi_kname = f"{_name}k{_props}"
            stochrsi_dname = f"{_name}d{_props}"
            df = talib.stochrsi(
                close,
                length=timeperiod,
                rsi_length=timeperiod,
                k=fastk_period,
                d=fastd_period,
                mamode=fastd_matype,
            )
            return df[stochrsi_kname], df[stochrsi_dname]
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.STOCHRSI(
                close.values, timeperiod, fastk_period, fastd_period, fastd_matype
            )

    @classmethod
    def highest(self, df,columnName, timeperiod):
        return df.rolling(timeperiod, min_periods=1)[columnName].max()

    @classmethod
    def lowest(self, df,columnName, timeperiod):
        return df.rolling(timeperiod, min_periods=1)[columnName].min()
    
    @classmethod
    def RVM(self, high, low, close, timeperiod):
        # Aligning the series
        # high,low,close = pktalib.align_series(high, low, close, fill_value=0)
        # Relative Volatality Measure
        #Short-term ATRs
        short1    = pktalib.ATR(high, low, close,3)
        short2    = pktalib.ATR(high, low, close,5)
        short3    = pktalib.ATR(high, low, close,8)
        shortAvg  = (short1 + short2 + short3) / 3

        #Long-term ATRs
        long1   = pktalib.ATR(high, low, close,55)
        long2   = pktalib.ATR(high, low, close,89)
        long3   = pktalib.ATR(high, low, close,144)
        longAvg = (long1 + long2 + long3) / 3

        #Combined ATR value
        combinedATR = (shortAvg + longAvg) / 2

        #Highest and lowest combined ATR over lookback period
        df_catr = pd.DataFrame(data=combinedATR,columns=["combinedATR"])
        highestCombinedATR = pktalib.highest(df_catr,"combinedATR", timeperiod)
        lowestCombinedATR = pktalib.lowest(df_catr,"combinedATR", timeperiod)

        #RVM Calculation
        diffLowest = (combinedATR - lowestCombinedATR)
        diffLowest = [x for x in diffLowest if ~np.isnan(x)]
        diffHighLow = (highestCombinedATR - lowestCombinedATR)
        diffHighLow = [x for x in diffHighLow if ~np.isnan(x)]
        df_diff_lowest = pd.DataFrame(data=diffLowest,columns=["diffLowest"])
        df_diff_highLow = pd.DataFrame(data=diffHighLow,columns=["diffHighLow"])
        maxHighLow = max(df_diff_highLow["diffHighLow"])
        rvm =  df_diff_lowest["diffLowest"]/ maxHighLow * 100
        return round(rvm.tail(1),1)

    @classmethod
    def ichimoku(
        self, df, tenkan=None, kijun=None, senkou=None, include_chikou=True, offset=None
    ):
        import pandas_ta_classic as ta

        ichimokudf, spandf = ta.ichimoku(
            df["high"], df["low"], df["close"], tenkan, kijun, senkou, False, 26
        )
        return ichimokudf

    @classmethod
    def supertrend(self, df, length=7, multiplier=3):
        import pandas_ta_classic as ta

        sti = ta.supertrend(
            df["high"], df["low"], df["close"], length=length, multiplier=multiplier
        )
        # trend, direction, long, short
        # SUPERT_7_3.0  SUPERTd_7_3.0  SUPERTl_7_3.0  SUPERTs_7_3.0
        return sti if sti is not None else {'SUPERT_7_3.0':np.nan}

    @classmethod
    def psar(self, high, low, acceleration=0.02, maximum=0.2):
        # Aligning the series
        # high,low = pktalib.align_series(high, low, fill_value=0)
        psar = talib.SAR(high, low, acceleration=acceleration, maximum=maximum)
        return psar

    # @classmethod
    # def momentum(self, df):
    #     df.loc[:,'MOM'] = talib.MOM(df.loc[:,"close"],2).apply(lambda x: round(x, 2))
    #     return df.loc[:,'MOM']

    # @classmethod
    # def get_dmi_df(self, df):
    #     df.loc[:,'DMI'] = talib.DX(df.loc[:,"high"],df.loc[:,"low"],df.loc[:,"close"],timeperiod=14)
    #     return df.loc[:,'DMI']

    # @classmethod
    # def get_macd_df(self, df):
    #     df.loc[:,'macd(12)'], df.loc[:,'macdsignal(9)'], df.loc[:,'macdhist(26)'] = talib.MACD(df.loc[:,"close"], fastperiod=12, slowperiod=26, signalperiod=9)
    #     df.loc[:,'macd(12)'] = df.loc[:,'macd(12)'].apply(lambda x: round(x, 3))
    #     df.loc[:,'macdsignal(9)']= df.loc[:,'macdsignal(9)'].apply(lambda x: round(x, 3))
    #     df.loc[:,'macdhist(26)'] = df.loc[:,'macdhist(26)'].apply(lambda x: round(x, 3))
    #     return df.loc[:,['macd(12)','macdsignal(9)', 'macdhist(26)']]

    # @classmethod
    # def get_sma_df(self, df):
    #     df.loc[:,'SMA(10)'] = talib.SMA(df.loc[:,"close"],10).apply(lambda x: round(x, 2))
    #     df.loc[:,'SMA(50)'] = talib.SMA(df.loc[:,"close"],50).apply(lambda x: round(x, 2))
    #     return df.loc[:,["close",'SMA(10)', 'SMA(50)']]

    # @classmethod
    # def get_ema_df(self, df):
    #     df.loc[:,'EMA(9)'] = talib.EMA(df.loc[:,"close"], timeperiod = 9).apply(lambda x: round(x, 2))
    #     return df.loc[:,["close",'EMA(9)']]

    # @classmethod
    # def get_adx_df(self, df):
    #     df.loc[:,'ADX'] = talib.ADX(df.loc[:,"high"],df.loc[:,"low"], df.loc[:,"close"], timeperiod=14).apply(lambda x: round(x, 2))
    #     return df.loc[:,'ADX']

    # @classmethod
    # def get_bbands_df(self, df):
    #     df.loc[:,'BBands-U'], df.loc[:,'BBands-M'], df.loc[:,'BBands-L'] = talib.BBANDS(df.loc[:,"close"], timeperiod =20)
    #     df.loc[:,'BBands-U'] = df.loc[:,'BBands-U'].apply(lambda x: round(x, 2))
    #     df.loc[:,'BBands-M'] = df.loc[:,'BBands-M'].apply(lambda x: round(x, 2))
    #     df.loc[:,'BBands-L'] = df.loc[:,'BBands-L'].apply(lambda x: round(x, 2))
    #     return df[["close",'BBands-U','BBands-M','BBands-L']]

    # @classmethod
    # def get_obv_df(self, df):
    #     if ("close" not in df.keys()) or ("volume" not in df.keys()):
    #         return np.nan
    #     df.loc[:,'OBV'] = talib.OBV(df.loc[:,"close"], df.loc[:,"volume"])
    #     return df.loc[:,'OBV']

    # @classmethod
    # def get_atr_df(self, df):
    #     df.loc[:,'ATR'] = talib.ATR(df.loc[:,"high"], df.loc[:,"low"], df.loc[:,"close"], timeperiod=14).apply(lambda x: round(x, 2))
    #     return df.loc[:,'ATR']

    # @classmethod
    # def get_natr_df(self, df):
    #     df.loc[:,'NATR'] = talib.NATR(df.loc[:,"high"], df.loc[:,"low"], df.loc[:,"close"], timeperiod=14).apply(lambda x: round(x, 2))
    #     return df.loc[:,'NATR']

    # @classmethod
    # def get_trange_df(self, df):
    #     df.loc[:,'TRANGE'] = talib.TRANGE(df.loc[:,"high"], df.loc[:,"low"], df.loc[:,"close"]).apply(lambda x: round(x, 2))
    #     return df.loc[:,'TRANGE']

    # @classmethod
    # def get_atr_extreme(self, df):
    #     """
    #     ATR Exterme: which is based on 《Volatility-Based Technical Analysis》
    #     TTI is 'Trading The Invisible'

    #     @return: fasts, slows
    #     """
    #     highs = df.loc[:,"high"]
    #     lows = df.loc[:,"low"]
    #     closes = df.loc[:,"close"]
    #     slowPeriod=30
    #     fastPeriod=3
    #     atr = self.get_atr_df(df)

    #     highsMean = talib.EMA(highs, 5)
    #     lowsMean = talib.EMA(lows, 5)
    #     closesMean = talib.EMA(closes, 5)

    #     atrExtremes = np.where(closes > closesMean,
    #                 ((highs - highsMean)/closes * 100) * (atr/closes * 100),
    #                 ((lows - lowsMean)/closes * 100) * (atr/closes * 100)
    #                 )
    #     fasts = talib.MA(atrExtremes, fastPeriod)
    #     slows = talib.EMA(atrExtremes, slowPeriod)
    #     return fasts, slows, np.std(atrExtremes[-slowPeriod:])

    # @classmethod
    # def get_atr_ratio(self, df):
    #     """
    #     ATR(14)/MA(14)
    #     """
    #     closes = df.loc[:,"close"]

    #     atr = self.get_atr_df(df)
    #     ma = talib.MA(closes, timeperiod=14)

    #     volatility = atr/ma

    #     s = pd.Series(volatility, index=df.index, name='volatility').dropna()
    #     pd.set_option('mode.chained_assignment', None)
    #     return pd.DataFrame({'volatility':round(s,2)})

    @classmethod
    def get_ppsr_df(self, high, low, close,pivotPoint=None):
        try:
            # Aligning the series
            # high,low,close = pktalib.align_series(high, low, close, fill_value=0)
            PSR = None
            if pivotPoint is None:
                return PSR
            PP = pd.Series((high + low + close) / 3)
            result = None
            if pivotPoint != "PP":
                if pivotPoint == "R1":
                    result = pd.Series(2 * PP - low)
                elif pivotPoint == "S1":
                    result = pd.Series(2 * PP - high)
                elif pivotPoint == "R2":
                    result = pd.Series(PP + high - low)
                elif pivotPoint == "S2":
                    result = pd.Series(PP - high + low)
                elif pivotPoint == "R3":
                    result = pd.Series(high + 2 * (PP - low))
                elif pivotPoint == "S3":
                    result = pd.Series(low - 2 * (high - PP))
            psr = {"close":close, 'PP':round(PP,2)}
            if pivotPoint != "PP" and result is not None:
                psr[pivotPoint] = round(result,2)
            with pd.option_context('mode.chained_assignment', None):
                PSR = pd.DataFrame(psr)
        except: # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            pass
        return PSR

    @classmethod
    def CDLMORNINGSTAR(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "morningstar")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDLMORNINGSTAR(open, high, low, close)

    @classmethod
    def CDLCUPANDHANDLE(self, open, high, low, close):
        if len(high) < 8:
            return False
        return (high.iloc[7] < high.iloc[6] and 
                high.iloc[7] < high.iloc[5] and 
                high.iloc[5] < high.iloc[4] and 
                high.iloc[5] < high.iloc[3] and 
                high.iloc[3] > high.iloc[2] and 
                high.iloc[0] > high.iloc[6])

    @classmethod
    def CDLMORNINGDOJISTAR(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "morningdojistar")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDLMORNINGDOJISTAR(open, high, low, close)

    @classmethod
    def CDLEVENINGSTAR(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "eveningstar")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDLEVENINGSTAR(open, high, low, close)

    @classmethod
    def CDLEVENINGDOJISTAR(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "eveningdojistar")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDLEVENINGDOJISTAR(open, high, low, close)

    @classmethod
    def CDLLADDERBOTTOM(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "ladderbottom")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDLLADDERBOTTOM(open, high, low, close)

    @classmethod
    def CDL3LINESTRIKE(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "3linestrike")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDL3LINESTRIKE(open, high, low, close)

    @classmethod
    def CDL3BLACKCROWS(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "3blackcrows")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDL3BLACKCROWS(open, high, low, close)

    @classmethod
    def CDL3INSIDE(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "3inside")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDL3INSIDE(open, high, low, close)

    @classmethod
    def CDL3OUTSIDE(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "3outside")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDL3OUTSIDE(open, high, low, close)

    @classmethod
    def CDL3WHITESOLDIERS(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "3whitesoldiers")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDL3WHITESOLDIERS(open, high, low, close)

    @classmethod
    def CDLHARAMI(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "harami")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDLHARAMI(open, high, low, close)

    @classmethod
    def CDLHARAMICROSS(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "haramicross")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDLHARAMICROSS(open, high, low, close)

    @classmethod
    def CDLMARUBOZU(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "marubozu")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDLMARUBOZU(open, high, low, close)

    @classmethod
    def CDLHANGINGMAN(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "hangingman")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDLHANGINGMAN(open, high, low, close)

    @classmethod
    def CDLHAMMER(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "hammer")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDLHAMMER(open, high, low, close)

    @classmethod
    def CDLINVERTEDHAMMER(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "invertedhammer")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDLINVERTEDHAMMER(open, high, low, close)

    @classmethod
    def CDLSHOOTINGSTAR(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "shootingstar")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDLSHOOTINGSTAR(open, high, low, close)

    @classmethod
    def CDLDRAGONFLYDOJI(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "dragonflydoji")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDLDRAGONFLYDOJI(open, high, low, close)

    @classmethod
    def CDLGRAVESTONEDOJI(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "gravestonedoji")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDLGRAVESTONEDOJI(open, high, low, close)

    @classmethod
    def CDLDOJI(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "doji")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDLDOJI(open, high, low, close)

    @classmethod
    def CDLENGULFING(self, open, high, low, close):
        try:
            # Aligning the series
            # open,high,low,close = pktalib.align_series(open,high, low, close, fill_value=0)
            return talib.cdl_pattern(open, high, low, close, "engulfing")
        except Exception:  # pragma: no cover
            # default_logger().debug(e, exc_info=True)
            return talib.CDLENGULFING(open, high, low, close)

    @classmethod
    def argrelextrema(self, data, comparator, axis=0, order=1, mode="clip"):
        """
        Calculate the relative extrema of `data`.

        Relative extrema are calculated by finding locations where
        ``comparator(data[n], data[n+1:n+order+1])`` is True.

        Parameters
        ----------
        data : ndarray
            Array in which to find the relative extrema.
        comparator : callable
            Function to use to compare two data points.
            Should take two arrays as arguments.
        axis : int, optional
            Axis over which to select from `data`. Default is 0.
        order : int, optional
            How many points on each side to use for the comparison
            to consider ``comparator(n,n+x)`` to be True.
        mode : str, optional
            How the edges of the vector are treated. 'wrap' (wrap around) or
            'clip' (treat overflow as the same as the last (or first) element).
            Default 'clip'. See numpy.take.

        Returns
        -------
        extrema : ndarray
            Boolean array of the same shape as `data` that is True at an extrema,
            False otherwise.

        See also
        --------
        argrelmax, argrelmin

        Examples
        --------
        >>> import numpy as np
        >>> testdata = np.array([1,2,3,2,1])
        >>> _boolrelextrema(testdata, np.greater, axis=0)
        array([False, False,  True, False, False], dtype=bool)

        """
        if (int(order) != order) or (order < 1):
            raise ValueError("Order must be an int >= 1")

        datalen = data.shape[axis]
        locs = np.arange(0, datalen)

        results = np.ones(data.shape, dtype=bool)
        main = data.take(locs, axis=axis, mode=mode)
        for shift in range(1, order + 1):
            plus = data.take(locs + shift, axis=axis, mode=mode)
            minus = data.take(locs - shift, axis=axis, mode=mode)
            results &= comparator(main, plus)
            results &= comparator(main, minus)
            if ~results.any():
                return results
        return np.nonzero(results)



"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import os
import sys
from PKNSETools.PKNSEStockDataFetcher import nseStockDataFetcher
from PKDevTools.classes.Singleton import SingletonType, SingletonMixin
from PKDevTools.classes.log import default_logger
from PKDevTools.classes.SuppressOutput import SuppressOutput
from PKDevTools.classes import log as log

class MarketStatus(SingletonMixin, metaclass=SingletonType):
    nseFetcher = nseStockDataFetcher()
    def __init__(self):
        super(MarketStatus, self).__init__()

    @property
    def exchange(self):
        if "exchange" in self.attributes.keys():
            return self.attributes["exchange"]
        else:
            return "^NSEI"
    
    @exchange.setter
    def exchange(self, exchangeKey):
        if self.exchange != exchangeKey:
            self.marketStatus = self.getMarketStatus(exchangeSymbol=exchangeKey)
        self.attributes["exchange"] = exchangeKey

    @property
    def marketStatus(self):
        if "marketStatus" in self.attributes.keys():
            return self.attributes["marketStatus"]
        else:
            # self.attributes["lock"] = "" # We don't need threading lock here
            self.marketStatus = ""
            return self.marketStatus
    
    @marketStatus.setter
    def marketStatus(self, status):
        self.attributes["marketStatus"] = status

    def getMarketStatus(self, progress=None, task_id=0, exchangeSymbol="^NSEI",namedOnly=False):
        return "NA"
        lngStatus = ""
        try:
            # if not 'pytest' in sys.modules:
            suppressLogs = True
            if "PKDevTools_Default_Log_Level" in os.environ.keys():
                suppressLogs = os.environ["PKDevTools_Default_Log_Level"] == str(log.logging.NOTSET)
            with SuppressOutput(suppress_stdout=suppressLogs, suppress_stderr=suppressLogs):
                if progress:
                    progress[task_id] = {"progress": 0, "total": 1}
                _,lngStatus,_ = "","TODO","" # MarketStatus.nseFetcher.capitalMarketStatus(exchange=exchangeSymbol)
                if exchangeSymbol in ["^NSEI","^BSESN"] and not namedOnly:
                    _,bseStatus,_ = "","TODO","" #MarketStatus.nseFetcher.capitalMarketStatus(exchange="^BSESN")
                    lngStatus = f"{lngStatus} | {bseStatus}"
            if progress:
                progress[task_id] = {"progress": 1, "total": 1}
        except KeyboardInterrupt: # pragma: no cover
            raise KeyboardInterrupt
        except Exception as e:# pragma: no cover
            default_logger().debug(e, exc_info=True)
            pass
        self.marketStatus = lngStatus
        return lngStatus



"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import sys
import os
from time import sleep
from enum import Enum

from PKDevTools.classes.Singleton import SingletonType, SingletonMixin
from pkscreener.classes.ConfigManager import tools, parser
from PKDevTools.classes.OutputControls import OutputControls
from PKDevTools.classes.ColorText import colorText
from PKDevTools.classes.Pikey import PKPikey
from PKDevTools.classes import Archiver
from PKDevTools.classes.log import default_logger
from pkscreener.classes import Utility, ConsoleUtility
from pkscreener.classes.MenuOptions import menus

class ValidationResult(Enum):
    Success = 0
    BadUserID = 1
    BadOTP = 2
    Trial = 3

class PKUserRegistration(SingletonMixin, metaclass=SingletonType):
    def __init__(self):
        super(PKUserRegistration, self).__init__()
        self._userID = 0
        self._otp = 0

    @classmethod
    def populateSavedUserCreds(self):
        configManager = tools()
        configManager.getConfig(parser)
        PKUserRegistration().userID = configManager.userID
        PKUserRegistration().otp = configManager.otp

    @property
    def userID(self):
        return self._userID

    @userID.setter
    def userID(self, newuserID):
        self._userID = newuserID

    @property
    def otp(self):
        return self._otp
    
    @otp.setter
    def otp(self, newotp):
        self._otp = newotp

    @classmethod
    def validateToken(self):
        try:
            if "RUNNER" in os.environ.keys():
                return True, ValidationResult.Success
            PKPikey.removeSavedFile(f"{PKUserRegistration().userID}")
            resp = Utility.tools.tryFetchFromServer(cache_file=f"{PKUserRegistration().userID}.pdf",directory="results/Data",hideOutput=True, branchName="SubData")
            if resp is None or resp.status_code != 200:
                return False, ValidationResult.BadUserID
            with open(os.path.join(Archiver.get_user_data_dir(),f"{PKUserRegistration().userID}.pdf"),"wb",) as f:
                f.write(resp.content)
            if not PKPikey.openFile(f"{PKUserRegistration().userID}.pdf",PKUserRegistration().otp):
                return False, ValidationResult.BadOTP
            return True, ValidationResult.Success
        except: # pragma: no cover
            if "RUNNER" in os.environ.keys():
                return True, ValidationResult.Success
            return False, ValidationResult.BadOTP

    @classmethod
    def login(self, trialCount=0):
        try:
            from pkscreener.classes.PKAnalytics import PKAnalyticsService
            PKAnalyticsService().collectMetrics()
            if "RUNNER" in os.environ.keys():
                return ValidationResult.Success
        except: # pragma: no cover
            return ValidationResult.BadUserID
        ConsoleUtility.PKConsoleTools.clearScreen(userArgs=None, clearAlways=True, forceTop=True)
        configManager = tools()
        configManager.getConfig(parser)
        if configManager.userID is not None and len(configManager.userID) > 0:
            PKUserRegistration.populateSavedUserCreds()
            if PKUserRegistration.validateToken()[0]:
                return ValidationResult.Success
        if trialCount >= 1:
            return PKUserRegistration.presentTrialOptions()

        OutputControls().printOutput(f"[+] {colorText.GREEN}PKScreener will always remain free and open source!{colorText.END}\n[+] {colorText.FAIL}PKScreener does offer certain premium/paid features!{colorText.END}\n[+] {colorText.GREEN}Please use {colorText.END}{colorText.WARN}@nse_pkscreener_bot{colorText.END}{colorText.GREEN} in telegram app on \n    your mobile phone to request your {colorText.END}{colorText.WARN}userID{colorText.END}{colorText.GREEN} and {colorText.END}{colorText.WARN}OTP{colorText.END}{colorText.GREEN} to login:\n{colorText.END}")
        username = None
        if configManager.userID is not None and len(configManager.userID) >= 1:
            username = input(f"[+] Your UserID from telegram: (Default: {colorText.GREEN}{configManager.userID}{colorText.END}): ") or configManager.userID
        else:
            username = input(f"[+] {colorText.GREEN}Your UserID from telegram: {colorText.END}")
        if username is None or len(username) <= 0:
            OutputControls().printOutput(f"{colorText.WARN}[+] We urge you to register on telegram (/OTP on @nse_pkscreener_bot) and then login to use PKScreener!{colorText.END}\n")
            OutputControls().printOutput(f"{colorText.FAIL}[+] Invalid userID!{colorText.END}\n{colorText.WARN}[+] Maybe try entering the {colorText.END}{colorText.GREEN}UserID{colorText.END}{colorText.WARN} instead of username?{colorText.END}\n[+] {colorText.WARN}If you have purchased a subscription and are still not able to login, please reach out to {colorText.END}{colorText.GREEN}@ItsOnlyPK{colorText.END} {colorText.WARN}on Telegram!{colorText.END}\n[+] {colorText.FAIL}Please try again or press Ctrl+C to exit!{colorText.END}")
            sleep(5)
            return PKUserRegistration.presentTrialOptions()
        otp = input(f"[+] {colorText.WARN}OTP received on telegram from {colorText.END}{colorText.GREEN}@nse_pkscreener_bot (Use command /otp to get OTP): {colorText.END}") or configManager.otp
        invalidOTP = False
        try:
            otpTest = int(otp)
        except KeyboardInterrupt: # pragma: no cover
            raise KeyboardInterrupt
        except Exception as e: # pragma: no cover
            default_logger().debug(e, exc_info=True)
            invalidOTP = True
            pass
        if otp is None or len(str(otp)) <= 0:
            OutputControls().printOutput(f"{colorText.WARN}[+] We urge you to register on telegram (/OTP on @nse_pkscreener_bot) and then login to use PKScreener!{colorText.END}\n")
            OutputControls().printOutput(f"{colorText.FAIL}[+] Invalid userID/OTP!{colorText.END}\n{colorText.WARN}[+] Maybe try entering the {colorText.END}{colorText.GREEN}UserID{colorText.END}{colorText.WARN} instead of username?{colorText.END}\n[+] {colorText.WARN}If you have purchased a subscription and are still not able to login, please reach out to {colorText.END}{colorText.GREEN}@ItsOnlyPK{colorText.END} {colorText.WARN}on Telegram!{colorText.END}\n[+] {colorText.FAIL}Please try again or press Ctrl+C to exit!{colorText.END}")
            sleep(5)
            return PKUserRegistration.presentTrialOptions()
    
        if len(str(otp)) <= 5 or invalidOTP:
            OutputControls().printOutput(f"{colorText.WARN}[+] Please enter a valid OTP!{colorText.END}\n[+] {colorText.FAIL}Please try again or press Ctrl+C to exit!{colorText.END}")
            sleep(3)
            return PKUserRegistration.login()
        try:
            userUsedUserID = False
            try:
                usernameInt = int(username)
                userUsedUserID = True
            except: # pragma: no cover
                userUsedUserID = False
                pass
            if userUsedUserID:
                OutputControls().printOutput(f"{colorText.GREEN}[+] Please wait!{colorText.END}\n[+] {colorText.WARN}Validating the OTP. You can press Ctrl+C to exit!{colorText.END}")
                PKUserRegistration().userID = usernameInt
                PKUserRegistration().otp = otp

                validationResult,validationReason = PKUserRegistration.validateToken()
                if not validationResult and validationReason == ValidationResult.BadUserID:
                    OutputControls().printOutput(f"{colorText.FAIL}[+] Invalid userID!{colorText.END}\n{colorText.WARN}[+] Maybe try entering the {colorText.END}{colorText.GREEN}UserID{colorText.END}{colorText.WARN} instead of username?{colorText.END}\n[+] {colorText.WARN}If you have purchased a subscription and are still not able to login, please reach out to {colorText.END}{colorText.GREEN}@ItsOnlyPK{colorText.END} {colorText.WARN}on Telegram!{colorText.END}\n[+] {colorText.FAIL}Please try again or press Ctrl+C to exit!{colorText.END}")
                    sleep(5)
                    return PKUserRegistration.presentTrialOptions()
                if not validationResult and validationReason == ValidationResult.BadOTP:
                    OutputControls().printOutput(f"{colorText.FAIL}[+] Invalid OTP!{colorText.END}\n[+] {colorText.GREEN}If you have purchased a subscription and are still not able to login, please reach out to @ItsOnlyPK on Telegram!{colorText.END}\n[+] {colorText.FAIL}Please try again or press Ctrl+C to exit!{colorText.END}")
                    sleep(5)
                    return PKUserRegistration.login(trialCount=trialCount+1)
                if validationResult and validationReason == ValidationResult.Success:
                    # Remember the userID for future login
                    configManager.userID = str(PKUserRegistration().userID)
                    configManager.otp = str(PKUserRegistration().otp)
                    configManager.setConfig(parser,default=True,showFileCreatedText=False)
                    ConsoleUtility.PKConsoleTools.clearScreen(userArgs=None, clearAlways=True, forceTop=True)
                    return validationReason
        except KeyboardInterrupt: # pragma: no cover
            raise KeyboardInterrupt
        except Exception as e: # pragma: n`o cover
            default_logger().debug(e, exc_info=True)
            pass
        OutputControls().printOutput(f"{colorText.WARN}[+] Invalid userID or OTP!{colorText.END}\n{colorText.GREEN}[+] May be try entering the {'UserID instead of username?' if userUsedUserID else 'Username instead of userID?'} {colorText.END}\n[+] {colorText.FAIL}Please try again or press Ctrl+C to exit!{colorText.END}")
        sleep(3)
        return PKUserRegistration.login(trialCount=trialCount+1)

    @classmethod
    def presentTrialOptions(self):
        m = menus()
        multilineOutputEnabled = OutputControls().enableMultipleLineOutput
        OutputControls().enableMultipleLineOutput = True
        m.renderUserType()
        userTypeOption = OutputControls().takeUserInput(colorText.FAIL + "  [+] Select option: ",enableUserInput=True,defaultInput="1")
        OutputControls().enableMultipleLineOutput = multilineOutputEnabled
        if str(userTypeOption).upper() in ["1"]:
            return PKUserRegistration.login(trialCount=0)
        elif str(userTypeOption).upper() in ["2"]:
            return ValidationResult.Trial
        sys.exit(0)
    


#!/usr/bin/python3
"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
# Pyinstaller compile Windows: pyinstaller --onefile --icon=screenshots\icon.ico pkscreener\pkscreenercli.py  --hidden-import cmath --hidden-import talib.stream --hidden-import numpy --hidden-import pandas --hidden-import alive_progress
# Pyinstaller compile Linux  : pyinstaller --onefile --icon=screenshots/icon.ico pkscreener/pkscreenercli.py  --hidden-import cmath --hidden-import talib.stream --hidden-import numpy --hidden-import pandas --hidden-import alive_progress
import warnings
warnings.simplefilter("ignore", UserWarning,append=True)
import argparse
import builtins
import logging
import json
import traceback
import datetime
# Keep module imports prior to classes
import os
import csv
import re
import sys
import tempfile
os.environ["PYTHONWARNINGS"]="ignore::UserWarning"
import multiprocessing

sys.path.append(os.path.dirname(os.path.abspath(__file__)))

try:
    logging.getLogger("tensorflow").setLevel(logging.ERROR)
except Exception:# pragma: no cover
    pass

from time import sleep
import time

from PKDevTools.classes import log as log
from PKDevTools.classes.ColorText import colorText
from PKDevTools.classes.log import default_logger
from PKDevTools.classes.PKDateUtilities import PKDateUtilities
from pkscreener import Imports
from PKDevTools.classes.OutputControls import OutputControls
from pkscreener.classes.MarketMonitor import MarketMonitor
import pkscreener.classes.ConfigManager as ConfigManager

if __name__ == '__main__':
    multiprocessing.freeze_support()
    # fix to https://stackoverflow.com/q/62748654/9191338
    # Python incorrectly tracks shared memory even if it is not
    # created by the process. The following patch is a workaround.
    from unittest.mock import patch
    patch("multiprocessing.resource_tracker.register",lambda *args, **kwargs: None)
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
os.environ["AUTOGRAPH_VERBOSITY"] = "0"

printenabled=False
originalStdOut=None
original__stdout=None
LoggedIn = False
cron_runs=0

def decorator(func):
    # @infer_global(new_func)
    def new_func(*args, **kwargs):
        if printenabled:
            try:
                func(*args,**kwargs)
            except Exception as e:# pragma: no cover
                default_logger().debug(e, exc_info=True)
                pass

    return new_func

# print = decorator(print) # current file
def disableSysOut(input=True, disable=True):
    global printenabled,originalStdOut, original__stdout
    printenabled = not disable
    if disable:
        if originalStdOut is None:
            builtins.print = decorator(builtins.print)  # all files
            if input:
                builtins.input = decorator(builtins.input)  # all files
            originalStdOut = sys.stdout
            original__stdout = sys.__stdout__
        sys.stdout = open(os.devnull, "w")
        sys.__stdout__ = open(os.devnull, "w")
    else:
        try:
            if originalStdOut is not None and original__stdout is not None:
                sys.stdout.close()
                sys.__stdout__.close()
        except Exception as e:# pragma: no cover
            default_logger().debug(e, exc_info=True)
            pass
        sys.stdout = originalStdOut if originalStdOut is not None else sys.stdout
        sys.__stdout__ = original__stdout if original__stdout is not None else sys.__stdout__

# Argument Parsing for test purpose
argParser = argparse.ArgumentParser()
argParser.add_argument(
    "-a",
    "--answerdefault",
    help="Pass default answer to questions/choices in the application. Example Y, N",
    required=False,
)
argParser.add_argument(
    "--backtestdaysago",
    help="Run scanner for -b days ago from today.",
    required=False,
)
argParser.add_argument(
    "--barometer",
    action="store_true",
    help="Send global market barometer to telegram channel or a user",
    required=False,
)
argParser.add_argument(
    "--bot",
    action="store_true",
    help="Run only in telegram bot mode",
    required=False,
)
argParser.add_argument(
    "--botavailable",
    action="store_true",
    help="Enforce whether bot is going to be available or not.",
    required=False,
)
argParser.add_argument(
    "-c",
    "--croninterval",
    help="Pass interval in seconds to wait before the program is run again with same parameters",
    required=False,
)
argParser.add_argument(
    "-d",
    "--download",
    action="store_true",
    help="Only download Stock data in .pkl file (No analysis will be run)",
    required=False,
)
argParser.add_argument(
    "-e",
    "--exit",
    action="store_true",
    help="Exit right after executing just once",
    required=False,
)
argParser.add_argument(
    "--fname",
    help="file name with results contents",
    required=False,
)
argParser.add_argument(
    "--forceBacktestsForZeroResultDays",
    help="Force run the backtests even for those days when we already have zero results saved in the repo",
    action=argparse.BooleanOptionalAction,
)
argParser.add_argument(
    "-i",
    "--intraday",
    help="Use Intraday configurations and use the candlestick duration that is passed. Acceptable values 1m, 5m, 10m, 15m, 1h etc.",
    required=False,
)
argParser.add_argument(
    "-m",
    "--monitor",
    help="Monitor for intraday scanners and their results.",
    nargs='?',
    const='X',
    type=str,
    required=False,
)
argParser.add_argument(
    "--maxdisplayresults",
    help="Maximum number of results to display.",
    required=False,
)
argParser.add_argument(
    "--maxprice",
    help="Maximum Price for the stock to be considered.",
    required=False,
)
argParser.add_argument(
    "--minprice",
    help="Minimum Price for the stock to be considered.",
    required=False,
)
argParser.add_argument(
    "-o",
    "--options",
    help="Pass selected options in the <MainMenu>:<SubMenu>:<SubMenu>:etc. format. For example: ./pkscreenercli.py -a Y -o X:12:10 -e will run the screener with answer Y as default choice to questions and scan with menu choices: Scanners > Nifty (All Stocks) > Closing at least 2%% up since last 3 day",
    required=False,
)
argParser.add_argument(
    "-p",
    "--prodbuild",
    action="store_true",
    help="Run in production-build mode",
    required=False,
)
argParser.add_argument(
    "--progressstatus",
    help="Pass default progress status that you'd like to get displayed when running the scans",
    required=False,
)
argParser.add_argument(
    "--runintradayanalysis",
    action="store_true",
    help="Run analysis for morning vs EoD LTP values",
    required=False,
)
argParser.add_argument(
    "--simulate",
    type=json.loads, # '{"isTrading":true,"currentDateTime":"2024-04-29 09:35:38"}'
    help="Simulate various conditions",
    required=False,
)
argParser.add_argument(
    "--singlethread",
    action="store_true",
    help="Run analysis for debugging purposes in a single process, single threaded environment",
    required=False,
)
argParser.add_argument(
    "--slicewindow",
    type=str,
    help="Time slice window value - a date or datetime string with timezone in international format",
    required=False,
)
argParser.add_argument(
    "--stocklist",
    type=str,
    help="Comma separated list of stocks passed from previous scan results",
    required=False,
)
argParser.add_argument(
    "--systemlaunched",
    action="store_true",
    help="Indicator to show that this is a system launched screener, using os.system",
    required=False,
)
argParser.add_argument(
    "-t",
    "--testbuild",
    action="store_true",
    help="Run in test-build mode",
    required=False,
)
argParser.add_argument(
    "--telegram",
    action="store_true",
    help="Run with an assumption that this instance is launched via telegram bot",
    required=False,
)
argParser.add_argument(
    "--triggertimestamp",
    help="Optionally, send the timestamp value when this was triggered",
    required=False,
)
argParser.add_argument(
    "-u",
    "--user",
    help="Telegram user ID to whom the results must be sent.",
    required=False,
)
argParser.add_argument(
    "-l",
    "--log",
    action="store_true",
    help="Run with full logging enabled",
    required=False,
)
argParser.add_argument("-v", action="store_true")  # Dummy Arg for pytest -v
argParser.add_argument(
    "--pipedtitle",
    help="Piped Titles",
    required=False,
)
argParser.add_argument(
    "--pipedmenus",
    help="Piped Menus",
    required=False,
)
argParser.add_argument(
    "--usertag",
    help="User defined tag value(s)",
    required=False,
)
argParser.add_argument(
    "--testalloptions",
    action="store_true",
    help="runs and tests all options",
    required=False,
)
from PKDevTools.classes.FunctionTimeouts import ping
from pkscreener.classes.PKAnalytics import PKAnalyticsService
def csv_split(s):
    return list(csv.reader([s], delimiter=' '))[0]

def re_split(s):
    def strip_quotes(s):
        if s and (s[0] == '"' or s[0] == "'") and s[0] == s[-1]:
            return s[1:-1]
        return s
    # pieces = [p for p in re.split("( |\\\".*?\\\"|'.*?')", s) if p.strip()]
    return [strip_quotes(p).replace('\\"', '"').replace("\\'", "'") for p in re.findall(r'(?:[^"\s]*"(?:\\.|[^"])*"[^"\s]*)+|(?:[^\'\s]*\'(?:\\.|[^\'])*\'[^\'\s]*)+|[^\s]+', s)]

def get_debug_args():
    global args
    try:
        if args is not None:
            # make sure that args are mutable
            args = list(args)
        return args
    except NameError as e: # pragma: no cover
        args = sys.argv[1:]
        if isinstance(args,list):
            if len(args) == 1:
                # re.findall(r'[^"\s]\S*|".+?"', line)
                # list(csv.reader([line], delimiter=" "))
                # pieces = [p for p in re.split("( |\\\".*?\\\"|'.*?')", test) if p.strip()]
                return re_split(args[0]) #args[0].split(" ")
            else:
                return args
        return None
    except TypeError as e: # pragma: no cover
        # NameSpace object is not iterable
        return args
    except Exception as e: # pragma: no cover
        return None
    # return ' --systemlaunched -a y -e -o "X:12:9:2.5:>|X:0:31:>|X:0:23:>|X:0:27:" -u -1001785195297 --stocklist GLS,NESCO,SBICARD,DREAMFOLKS,JAGRAN,ACEINTEG,RAMASTEEL'.split(" ")

args = get_debug_args()
argsv = argParser.parse_known_args(args=args)
# argsv = argParser.parse_known_args()
args = argsv[0]
# args.slicewindow = "2024-09-06 10:55:12.481253+05:30"
results = None
resultStocks = None
plainResults = None
start_time = None
dbTimestamp = None
elapsed_time = None
configManager = ConfigManager.tools()


def exitGracefully():
    try:
        from PKDevTools.classes import Archiver
        from pkscreener.globals import resetConfigToDefault
        filePath = None
        try:
            filePath = os.path.join(Archiver.get_user_data_dir(), "monitor_outputs")
        except: # pragma: no cover
            pass
        if filePath is None:
            return
        index = 0
        while index < configManager.maxDashboardWidgetsPerRow*configManager.maxNumResultRowsInMonitor:
            try:
                os.remove(f"{filePath}_{index}.txt")
            except: # pragma: no cover
                pass
            index += 1

        argsv = argParser.parse_known_args()
        args = argsv[0]
        if args is not None and args.options is not None and not args.options.upper().startswith("T"):
            resetConfigToDefault(force=True)
            
        if "PKDevTools_Default_Log_Level" in os.environ.keys():
            if args is None or (args is not None and args.options is not None and "|" not in args.options):
                del os.environ['PKDevTools_Default_Log_Level']
        configManager.logsEnabled = False
        configManager.setConfig(ConfigManager.parser,default=True,showFileCreatedText=False)
    except RuntimeError: # pragma: no cover
        OutputControls().printOutput(f"{colorText.WARN}If you're running from within docker, please run like this:{colorText.END}\n{colorText.FAIL}docker run -it pkjmesra/pkscreener:latest\n{colorText.END}")
        pass

def logFilePath():
    try:
        from PKDevTools.classes import Archiver

        filePath = os.path.join(Archiver.get_user_data_dir(), "pkscreener-logs.txt")
        f = open(filePath, "w")
        if f is not None:
            f.write("Logger file for pkscreener!")
            f.close()
    except Exception:# pragma: no cover
        filePath = os.path.join(tempfile.gettempdir(), "pkscreener-logs.txt")
    return filePath


def setupLogger(shouldLog=False, trace=False):
    if not shouldLog:
        if "PKDevTools_Default_Log_Level" in os.environ.keys():
            del os.environ['PKDevTools_Default_Log_Level']
        return
    log_file_path = logFilePath()

    if os.path.exists(log_file_path):
        try:
            os.remove(log_file_path)
        except Exception:# pragma: no cover
            pass
    OutputControls().printOutput(colorText.FAIL + "\n  [+] Logs will be written to:"+colorText.END)
    OutputControls().printOutput(colorText.GREEN + f"  [+] {log_file_path}"+colorText.END)
    OutputControls().printOutput(colorText.FAIL + "  [+] If you need to share, open this folder, copy and zip the log file to share.\n" + colorText.END)
    # logger = multiprocessing.log_to_stderr(log.logging.DEBUG)
    os.environ["PKDevTools_Default_Log_Level"] = str(log.logging.DEBUG)
    log.setup_custom_logger(
        "pkscreener",
        log.logging.DEBUG,
        trace=trace,
        log_file_path=log_file_path,
        filter=None,
    )

def warnAboutDependencies():
    if not Imports["talib"]:
        OutputControls().printOutput(
                colorText.FAIL
                + "  [+] TA-Lib is not installed. Looking for pandas_ta_classic."
                + colorText.END
            )
        sleep(1)
        issueLink = "https://github.com/pkjmesra/PKScreener"
        issueLink = f"\x1b[97m\x1b]8;;{issueLink}\x1b\\{issueLink}\x1b]8;;\x1b\\\x1b[0m"
        if Imports["pandas_ta_classic"]:
            taLink = "https://github.com/ta-lib/ta-lib-python"
            taLink = f"\x1b[97m\x1b]8;;{taLink}\x1b\\{taLink}\x1b]8;;\x1b\\\x1b[0m"
            OutputControls().printOutput(
                colorText.GREEN
                + f"  [+] Found and falling back on pandas_ta_classic.\n  [+] For full coverage(candle patterns), you may wish to read the README file in PKScreener repo :  {issueLink}\n  [+] or follow instructions from\n  [+] {taLink}"
                + colorText.END
            )
            sleep(1)
        else:
            OutputControls().printOutput(
                colorText.FAIL
                + f"  [+] Neither ta-lib nor pandas_ta_classic was located. You need at least one of them to continue! \n  [+] Please follow instructions from README file under PKScreener repo: {issueLink}"
                + colorText.END
            )
            OutputControls().takeUserInput("Press any key to try anyway...")
    
def runApplication():
    from pkscreener.globals import main, sendQuickScanResult,sendMessageToTelegramChannel, sendGlobalMarketBarometer, updateMenuChoiceHierarchy, isInterrupted, refreshStockData, closeWorkersAndExit, resetUserMenuChoiceOptions,menuChoiceHierarchy
    # From a previous call to main with args, it may have been mutated.
    # Let's stock to the original args passed by user
    try:
        savedPipedArgs = None
        savedPipedArgs = args.pipedmenus if args is not None and args.pipedmenus is not None else None
    except: # pragma: no cover
        pass
    global results, resultStocks, plainResults, dbTimestamp, elapsed_time, start_time,argParser
    from pkscreener.classes.MenuOptions import menus, PREDEFINED_SCAN_MENU_TEXTS, PREDEFINED_PIPED_MENU_ANALYSIS_OPTIONS,PREDEFINED_SCAN_MENU_VALUES
    args = get_debug_args()
    monitorOption = None
    if not isinstance(args,argparse.Namespace) and not hasattr(args, "side_effect"):
        argsv = argParser.parse_known_args(args=args)
        # argsv = argParser.parse_known_args()
        args = argsv[0]
    # During the previous run, we may have made several changes to the original args variable.
    # We need to re-load the original args
    if args is not None and not args.exit and not args.monitor:
        argsv = argParser.parse_known_args()
        args = argsv[0]
    # args.slicewindow = "2024-09-06 10:55:12.481253+05:30"
    if args.user is None:
        from PKDevTools.classes.Environment import PKEnvironment
        Channel_Id, _, _, _ = PKEnvironment().secrets
        if Channel_Id is not None and len(str(Channel_Id)) > 0:
            args.user = int(f"-{Channel_Id}")
    if args.triggertimestamp is None:
        args.triggertimestamp = int(PKDateUtilities.currentDateTimestamp())
    else:
        args.triggertimestamp = int(args.triggertimestamp)
    if args.systemlaunched and args.options is not None:
        args.systemlaunched = args.options
    
    # if sys.argv[0].endswith(".py"):
    #     args.monitor = 'X'
    #     args.answerdefault = 'Y'
    args.pipedmenus = savedPipedArgs
    if args.options is not None:
        args.options = args.options.replace("::",":").replace("\"","").replace("'","")
        if args.options.upper().startswith("C") or "C:" in args.options.upper():
            args.runintradayanalysis = True
        args,choices = updateProgressStatus(args)
        
    if args.runintradayanalysis:
        generateIntradayAnalysisReports(args)
    else:
        if args.testalloptions:
            allMenus,_ = menus.allMenus(index=0)
            for scanOption in allMenus:
                 args.options = f"{scanOption}:SBIN,"
                 _, _ = main(userArgs=args)
            sys.exit(0)

        if args.barometer:
            sendGlobalMarketBarometer(userArgs=args)
            sys.exit(0)
        else:
            monitorOption_org = ""
            # args.monitor = configManager.defaultMonitorOptions
            if args.monitor:
                args.monitor = args.monitor.replace("::",":").replace("\"","").replace("'","")
                configManager.getConfig(ConfigManager.parser)
                args.answerdefault = args.answerdefault or 'Y'
                MarketMonitor().hiddenColumns = configManager.alwaysHiddenDisplayColumns
                if MarketMonitor().monitorIndex == 0:
                    dbTimestamp = PKDateUtilities.currentDateTime().strftime("%H:%M:%S")
                    elapsed_time = 0
                    if start_time is None:
                        start_time = time.time()
                    else:
                        elapsed_time = round(time.time() - start_time,2)
                        start_time = time.time()
                monitorOption_org = MarketMonitor().currentMonitorOption()
                monitorOption = monitorOption_org.replace("::",":").replace("\"","").replace("'","")
                monitorOption = checkIntradayComponent(args, monitorOption)
                if monitorOption.startswith("|"):
                    monitorOption = monitorOption[1:]
                    monitorOptions = monitorOption.split(":")
                    if monitorOptions[1] != "0":
                        monitorOptions[1] = "0"
                        monitorOption = ":".join(monitorOptions)
                    # We need to pipe the output from previous run into the next one
                    if monitorOption.startswith("{") and "}" in monitorOption:
                        srcIndex = monitorOption.split("}")[0].split("{")[-1]
                        monitorOption="".join(monitorOption.split("}")[1:])
                        try:
                            srcIndex = int(srcIndex)
                            # Let's get the previously saved result for the monitor
                            savedStocks = MarketMonitor().monitorResultStocks[str(srcIndex)]
                            innerPipes = monitorOption.split("|")
                            nextPipe = innerPipes[0]
                            nextMonitor = nextPipe.split(">")[0]
                            innerPipes[0] = f"{nextMonitor}:{savedStocks}"
                            monitorOption = ":>|".join(innerPipes)
                            monitorOption = monitorOption.replace("::",":").replace(":>:>",":>")
                            # monitorOption = f"{monitorOption}:{savedStocks}:"
                        except: # pragma: no cover
                            # Probably wrong (non-integer) index passed. Let's continue anyway
                            pass
                    elif resultStocks is not None:
                        resultStocks = ",".join(resultStocks)
                        monitorOption = f"{monitorOption}:{resultStocks}"
                args.options = monitorOption.replace("::",":")
                fullMonitorMode = MarketMonitor().monitorIndex == 1 and args.options is not None and plainResults is not None
                partMonitorMode = len(MarketMonitor().monitors) == 1 and args.options is not None and plainResults is not None
                if (fullMonitorMode or partMonitorMode):
                    # Load the stock data afresh for each cycle
                    refreshStockData(args.options)
            try:
                results = None
                plainResults = None
                resultStocks = None
                if args is not None and ((args.options is not None and "|" in args.options) or args.systemlaunched):
                    args.maxdisplayresults = 2000
                updateConfigDurations(args=args)
                updateConfig(args=args)
                results, plainResults = main(userArgs=args)
                if args.pipedmenus is not None:
                    while args.pipedmenus is not None:
                        args,_ = updateProgressStatus(args,monitorOptions=monitorOption)
                        results, plainResults = main(userArgs=args)
                    # sys.exit(0)
                if isInterrupted():
                    closeWorkersAndExit()
                    exitGracefully()
                    sys.exit(0)
                runPipedScans = True
                while runPipedScans:
                    runPipedScans = pipeResults(plainResults,args)
                    if runPipedScans:
                        args,_ = updateProgressStatus(args,monitorOptions=monitorOption)
                        results, plainResults = main(userArgs=args)
                    else:
                        if args is not None and args.pipedtitle is not None and "|" in args.pipedtitle:
                            OutputControls().printOutput(
                                    colorText.WARN
                                    + f"  [+] Pipe Results Found: {args.pipedtitle}. {'Reduce number of piped scans if no stocks could be found.' if '[0]' in args.pipedtitle else ''}"
                                    + colorText.END
                                )
                            if args.answerdefault is None:
                                OutputControls().takeUserInput("Press <Enter> to continue...")
            except SystemExit: # pragma: no cover
                closeWorkersAndExit()
                exitGracefully()
                sys.exit(0)
            except KeyboardInterrupt: # pragma: no cover
                closeWorkersAndExit()
                exitGracefully()
                sys.exit(0)
            except Exception as e: # pragma: no cover
                default_logger().debug(e, exc_info=True)
                if args.log:
                    traceback.print_exc()
                # Probably user cancelled an operation by choosing a cancel sub-menu somewhere
                pass
            if plainResults is not None and not plainResults.empty:
                try:
                    plainResults.set_index("Stock", inplace=True)
                except: # pragma: no cover
                    pass
                try:
                    results.set_index("Stock", inplace=True)
                except: # pragma: no cover
                    pass
                plainResults = plainResults[~plainResults.index.duplicated(keep='first')]
                results = results[~results.index.duplicated(keep='first')]
                resultStocks = plainResults.index
            if args.monitor is not None:
                MarketMonitor().saveMonitorResultStocks(plainResults)
                if results is not None and len(monitorOption_org) > 0:
                    chosenMenu = args.pipedtitle if args.pipedtitle is not None else updateMenuChoiceHierarchy()
                    MarketMonitor().refresh(screen_df=results,screenOptions=monitorOption_org, chosenMenu=chosenMenu[:120],dbTimestamp=f"{dbTimestamp} | CycleTime:{elapsed_time}s",telegram=args.telegram)
                    menuChoiceHierarchy = ""
                    args.pipedtitle = ""
                # check to see if the monitor was launched before the market close hours.
                # If so, close it.
                if "RUNNER" in os.environ.keys() and args.triggertimestamp is not None:
                    from datetime import timezone
                    from PKDevTools.classes.MarketHours import MarketHours
                    marketCloseTS = PKDateUtilities.currentDateTime(simulate=True,hour=MarketHours().closeHour,minute=MarketHours().closeMinute).replace(tzinfo=timezone.utc).timestamp()
                    if int(args.triggertimestamp) < int(marketCloseTS) and int(PKDateUtilities.currentDateTimestamp()) >= marketCloseTS:
                        OutputControls().printOutput("Exiting monitor now since market has closed!",enableMultipleLineOutput=True)
                        sys.exit(0)

def updateProgressStatus(args,monitorOptions=None):
    from pkscreener.classes.MenuOptions import PREDEFINED_SCAN_MENU_TEXTS,PREDEFINED_SCAN_MENU_VALUES
    try:
        choices = ""
        if args.systemlaunched or monitorOptions is not None:
            optionsToUse = args.options if monitorOptions is None else monitorOptions
            choices = f"--systemlaunched -a y -e -o '{optionsToUse.replace('C:','X:').replace('D:','')}'"
            from pkscreener.classes.MenuOptions import INDICES_MAP
            searchChoices = choices
            for indexKey in INDICES_MAP.keys():
                if indexKey.isnumeric():
                    searchChoices = searchChoices.replace(f"X:{indexKey}:","X:12:")
            indexNum = PREDEFINED_SCAN_MENU_VALUES.index(searchChoices)
            selectedIndexOption = choices.split(":")[1]
            choices = f"P_1_{str(indexNum +1)}_{str(selectedIndexOption)}" if ">|" in choices else choices
            args.progressstatus = f"  [+] {choices} => Running {choices}"
            args.usertag = PREDEFINED_SCAN_MENU_TEXTS[indexNum]
            args.maxdisplayresults = 2000 #if monitorOptions is None else 100
    except: # pragma: no cover
        choices = ""
        pass
    return args, choices

def generateIntradayAnalysisReports(args):
    from pkscreener.globals import main, isInterrupted, closeWorkersAndExit, resetUserMenuChoiceOptions
    from pkscreener.classes.MenuOptions import menus, PREDEFINED_SCAN_MENU_TEXTS, PREDEFINED_PIPED_MENU_ANALYSIS_OPTIONS,PREDEFINED_SCAN_MENU_VALUES
    from PKDevTools.classes import Archiver
    maxdisplayresults = configManager.maxdisplayresults
    configManager.maxdisplayresults = 2000
    configManager.setConfig(ConfigManager.parser, default=True, showFileCreatedText=False)
    runOptions = []
    otherMenus = []
    if len(args.options.split(":")) >= 4:
        runOptions = [args.options]
    else:
        runOptions = PREDEFINED_PIPED_MENU_ANALYSIS_OPTIONS
            # otherMenus =  menus.allMenus(topLevel="C", index=12)
        if len(otherMenus) > 0:
            runOptions.extend(otherMenus)
    import pandas as pd
    optionalFinalOutcome_df = pd.DataFrame()
    from pkscreener.classes import Utility, ConsoleUtility
    # Delete any existing data from the previous run.
    configManager.deleteFileWithPattern(rootDir=Archiver.get_user_data_dir(),pattern="stock_data_*.pkl")
    analysis_index = 1
    for runOption in runOptions:
        try:
            runOptionName = f"--systemlaunched -a y -e -o '{runOption.replace('C:','X:').replace('D:','')}'"
            indexNum = PREDEFINED_SCAN_MENU_VALUES.index(runOptionName)
            runOptionName = f"{'  [+] P_1_'+str(indexNum +1) if '>|' in runOption else runOption}"
        except Exception as e: # pragma: no cover
            default_logger().debug(e,exc_info=True)
            runOptionName = f"  [+] {runOption.replace('D:','').replace(':D','').replace(':','_').replace('_D','').replace('C_','X_')}"
            pass
        args.progressstatus = f"{runOptionName} => Running Intraday Analysis: {analysis_index} of {len(runOptions)}..."
        analysisOptions = runOption.split("|")
        analysisOptions[-1] = analysisOptions[-1].replace("X:","C:")
        runOption = "|".join(analysisOptions)
        args.options = runOption
        try:
            results,plainResults = main(userArgs=args,optionalFinalOutcome_df=optionalFinalOutcome_df)
            if args.pipedmenus is not None:
                while args.pipedmenus is not None:
                    results, plainResults = main(userArgs=args)
            if isInterrupted():
                closeWorkersAndExit()
                exitGracefully()
                sys.exit(0)
            runPipedScans = True
            while runPipedScans:
                runPipedScans = pipeResults(plainResults,args)
                if runPipedScans:
                    results, plainResults = main(userArgs=args,optionalFinalOutcome_df=optionalFinalOutcome_df)
            if results is not None and len(results) >= len(optionalFinalOutcome_df) and not results.empty and len(results.columns) > 5:
                import numpy as np
                if "%Chng" in results.columns and "EoDDiff" in results.columns:
                    optionalFinalOutcome_df = results
            if optionalFinalOutcome_df is not None and "EoDDiff" not in optionalFinalOutcome_df.columns:
                # Somehow the file must have been corrupted. Let's re-download
                configManager.deleteFileWithPattern(rootDir=Archiver.get_user_data_dir(), pattern="*stock_data_*.pkl")
                configManager.deleteFileWithPattern(rootDir=Archiver.get_user_data_dir(), pattern="*intraday_stock_data_*.pkl")
            if isInterrupted():
                break
        except KeyboardInterrupt: # pragma: no cover
            closeWorkersAndExit()
            exitGracefully()
            sys.exit(0)
        except Exception as e: # pragma: no cover
            OutputControls().printOutput(e)
            if args.log:
                traceback.print_exc()
        resetUserMenuChoiceOptions()
        analysis_index += 1
            # saveSendFinalOutcomeDataframe(optionalFinalOutcome_df)

    configManager.maxdisplayresults = maxdisplayresults
    configManager.setConfig(ConfigManager.parser, default=True, showFileCreatedText=False)
    saveSendFinalOutcomeDataframe(optionalFinalOutcome_df)

def saveSendFinalOutcomeDataframe(optionalFinalOutcome_df):
    import pandas as pd
    from pkscreener.classes import Utility, ConsoleUtility
    from pkscreener.globals import sendQuickScanResult,showBacktestResults

    if optionalFinalOutcome_df is not None and not optionalFinalOutcome_df.empty:
        final_df = None
        try:
            optionalFinalOutcome_df.drop('FairValue', axis=1, inplace=True, errors="ignore")
            df_grouped = optionalFinalOutcome_df.groupby("Stock")
            for stock, df_group in df_grouped:
                if stock == "BASKET":
                    if final_df is None:
                        final_df = df_group[["Pattern","LTP","LTP@Alert","SqrOffLTP","SqrOffDiff","EoDDiff","DayHigh","DayHighDiff"]]
                    else:
                        final_df = pd.concat([final_df, df_group[["Pattern","LTP","LTP@Alert","SqrOffLTP","SqrOffDiff","EoDDiff","DayHigh","DayHighDiff"]]], axis=0)
        except: # pragma: no cover
            pass
        if final_df is not None and not final_df.empty:
            with pd.option_context('mode.chained_assignment', None):
                final_df = final_df[["Pattern","LTP@Alert","LTP","EoDDiff","DayHigh","DayHighDiff"]] # "SqrOffLTP","SqrOffDiff"
                final_df.rename(
                        columns={
                            "Pattern": "Scan Name",
                            "LTP@Alert": "Basket Value@Alert",
                            "LTP": "Basket Value@EOD",
                            # "SqrOffLTP": "Basket Value@SqrOff",
                            "DayHigh": "Basket Value@DayHigh",
                            },
                            inplace=True,
                        )
                final_df.dropna(inplace=True)
                final_df.dropna(how= "all", axis=1, inplace=True)
            mark_down = colorText.miniTabulator().tabulate(
                                    final_df,
                                    headers="keys",
                                    tablefmt=colorText.No_Pad_GridFormat,
                                    showindex = False
                                ).encode("utf-8").decode(Utility.STD_ENCODING)
            showBacktestResults(final_df,optionalName="Intraday_Backtest_Result_Summary",choices="Summary")
            OutputControls().printOutput(mark_down)
            from PKDevTools.classes.Environment import PKEnvironment
            Channel_Id, _, _, _ = PKEnvironment().secrets
            if Channel_Id is not None and len(str(Channel_Id)) > 0:
                sendQuickScanResult(menuChoiceHierarchy="IntradayAnalysis (If you would have bought at alert time and sold at end of day or day high)",
                                        user=int(f"-{Channel_Id}"),
                                        tabulated_results=mark_down,
                                        markdown_results=mark_down,
                                        caption="Intraday Analysis Summary - Morning alert vs Market Close",
                                        pngName= f"PKS_IA_{PKDateUtilities.currentDateTime().strftime('%Y-%m-%d_%H:%M:%S')}",
                                        pngExtension= ".png",
                                        forceSend=True
                                        )

def checkIntradayComponent(args, monitorOption):
    lastComponent = monitorOption.split(":")[-1]
    if "i" not in lastComponent:
        possiblePositions = monitorOption.split(":i")
        if len(possiblePositions) > 1:
            lastComponent = f"i {possiblePositions[1]}"
                # previousCandleDuration = configManager.duration
    if "i" in lastComponent:
        # We need to switch to intraday scan
        monitorOption = monitorOption.replace(lastComponent,"")
        args.intraday = lastComponent.replace("i","").strip()
        configManager.toggleConfig(candleDuration=args.intraday, clearCache=False)
        # args.options = f"{monitorOption}:{args.options[len(lastComponent):]}"
    else:
        # We need to switch to daily scan
        args.intraday = None
        configManager.toggleConfig(candleDuration='1d', clearCache=False)
    return monitorOption

def updateConfigDurations(args):
    if args is None or args.options is None:
        return
    nextOnes = args.options.split(">")
    if len(nextOnes) > 1:
        monitorOption = nextOnes[0]
        if len(monitorOption) == 0:
            return
        lastComponent = ":".join(monitorOption.split(":")[-2:])
        if "i" in lastComponent and "," not in lastComponent and " " in lastComponent:
            if "i" in lastComponent.split(":")[-2]:
                lastComponent = lastComponent.split(":")[-2]
            else:
                lastComponent = lastComponent.split(":")[-1]
            # We need to switch to intraday scan
            args.intraday = lastComponent.replace("i","").strip()
            configManager.toggleConfig(candleDuration=args.intraday, clearCache=False)
        else:
            # We need to switch to daily scan
            args.intraday = None
            configManager.toggleConfig(candleDuration='1d', clearCache=False)

def pipeResults(prevOutput,args):
    if args is None or args.options is None:
        return False
    hasFoundStocks = False
    nextOnes = args.options.split(">")
    if len(nextOnes) > 1:
        monitorOption = nextOnes[1]
        if len(monitorOption) == 0:
            return False
        lastComponent = ":".join(monitorOption.split(":")[-2:])
        if "i" in lastComponent and "," not in lastComponent and " " in lastComponent:
            if "i" in lastComponent.split(":")[-2]:
                lastComponent = lastComponent.split(":")[-2]
            else:
                lastComponent = lastComponent.split(":")[-1]
            # We need to switch to intraday scan
            monitorOption = monitorOption.replace(lastComponent,"")
            args.intraday = lastComponent.replace("i","").strip()
            configManager.toggleConfig(candleDuration=args.intraday, clearCache=False)
        else:
            # We need to switch to daily scan
            args.intraday = None
            configManager.toggleConfig(candleDuration='1d', clearCache=False)
        if monitorOption.startswith("|"):
            monitorOption = monitorOption.replace("|","")
            monitorOptions = monitorOption.split(":")
            if monitorOptions[0].upper() in ["X","C"] and monitorOptions[1] != "0":
                monitorOptions[1] = "0"
                monitorOption = ":".join(monitorOptions)
            if "B" in monitorOptions[0].upper() and monitorOptions[1] != "30":
                monitorOption = ":".join(monitorOptions).upper().replace(f"{monitorOptions[0].upper()}:{monitorOptions[1]}",f"{monitorOptions[0].upper()}:30:{monitorOptions[1]}")
            # We need to pipe the output from previous run into the next one
            if prevOutput is not None and not prevOutput.empty:
                try:
                    prevOutput.set_index("Stock", inplace=True)
                except: # pragma: no cover
                    pass
                prevOutput_results = prevOutput[~prevOutput.index.duplicated(keep='first')]
                prevOutput_results = prevOutput_results.index
                hasFoundStocks = len(prevOutput_results) > 0
                prevOutput_results = ",".join(prevOutput_results)
                monitorOption = monitorOption.replace(":D:",":")
                monitorOption = f"{monitorOption}:{prevOutput_results}"
        args.options = monitorOption.replace("::",":")
        args.options = args.options + ":D:>" + ":D:>".join(nextOnes[2:])
        args.options = args.options.replace("::",":")
        return True and hasFoundStocks
    return False

def removeOldInstances():
    import glob
    pattern = "pkscreenercli*"
    thisInstance = sys.argv[0]
    for f in glob.glob(pattern, root_dir=os.getcwd(), recursive=True):
        fileToDelete = f if (os.sep in f and f.startswith(thisInstance[:10])) else os.path.join(os.getcwd(),f)
        if not fileToDelete.endswith(thisInstance):
            try:
                os.remove(fileToDelete)
            except: # pragma: no cover
                pass

def updateConfig(args):
    if args is None:
        return
    configManager.getConfig(ConfigManager.parser)
    if args.intraday:
        configManager.toggleConfig(candleDuration=args.intraday, clearCache=False)
        if configManager.candlePeriodFrequency not in ["d","mo"] or configManager.candleDurationFrequency not in ["m"]:
            configManager.period = "1d"
            configManager.duration = args.intraday
            configManager.setConfig(ConfigManager.parser,default=True, showFileCreatedText=False)
    elif configManager.candlePeriodFrequency not in ["y","max","mo"] or configManager.candleDurationFrequency not in ["d","wk","mo","h"]:
        if args.answerdefault is not None or args.systemlaunched:
            configManager.period = "1y"
            configManager.duration = "1d"
            configManager.setConfig(ConfigManager.parser,default=True, showFileCreatedText=False)

@ping(interval=60,instance=PKAnalyticsService())
def pkscreenercli():
    global originalStdOut, args
    if sys.platform.startswith("darwin"):
        try:
            multiprocessing.set_start_method("fork")
        except RuntimeError as e:# pragma: no cover
            if "RUNNER" not in os.environ.keys() and ('PKDevTools_Default_Log_Level' in os.environ.keys() and os.environ["PKDevTools_Default_Log_Level"] != str(log.logging.NOTSET)):
                OutputControls().printOutput(
                    "  [+] RuntimeError with 'multiprocessing'.\n  [+] Please contact the Developer, if this does not work!"
                )
                OutputControls().printOutput(e)
                traceback.print_exc()
            pass
        # finally:
        #     from PKDevTools.classes.PKBackupRestore import restore_backup
        #     restore_backup()
        #     sleep(3)
        #     import threading
        #     from pkscreener.globals import tryLoadDataOnBackgroundThread
        #     ping_thread = threading.Thread(target=tryLoadDataOnBackgroundThread, daemon=True)
        #     ping_thread.start()
    try:
        removeOldInstances()
        OutputControls(enableMultipleLineOutput=(args is None or args.monitor is None or args.runintradayanalysis),enableUserInput=(args is None or args.answerdefault is None)).printOutput("",end="\r")
        configManager.getConfig(ConfigManager.parser)
        userAcceptance = configManager.tosAccepted
        inContainer = os.environ.get("PKSCREENER_DOCKER", "").lower() in ("yes", "y", "on", "true", "1")
        if not configManager.tosAccepted:
            if (args is not None and args.answerdefault is not None and str(args.answerdefault).lower() == "n"):
                OutputControls().printOutput(f"{colorText.FAIL}You seem to have passed disagreement to the Disclaimer and Terms Of Service of PKScreener by passing in {colorText.END}{colorText.WARN}--answerdefault N or -a N{colorText.END}. Exiting now!")
                sleep(5)
                sys.exit(0)
            allArgs = args.__dict__
            disclaimerLink = '\x1b[97m\x1b]8;;https://pkjmesra.github.io/PKScreener/Disclaimer.txt\x1b\\https://pkjmesra.github.io/PKScreener/Disclaimer.txt\x1b]8;;\x1b\\\x1b[0m'
            tosLink = '\x1b[97m\x1b]8;;https://pkjmesra.github.io/PKScreener/tos.txt\x1b\\https://pkjmesra.github.io/PKScreener/tos.txt\x1b]8;;\x1b\\\x1b[0m'
            for argKey in allArgs.keys():
                arg = allArgs[argKey]
                if arg is not None and arg:
                    userAcceptance = True
                    OutputControls().printOutput(f"{colorText.GREEN}By using this Software and passing a value for [{argKey}={arg}], you agree to\n[+] having read through the Disclaimer{colorText.END} ({disclaimerLink})\n[+]{colorText.GREEN} and accept Terms Of Service {colorText.END}({tosLink}){colorText.GREEN} of PKScreener. {colorText.END}\n[+] {colorText.WARN}If that is not the case, you MUST immediately terminate PKScreener by pressing Ctrl+C now!{colorText.END}")
                    sleep(2)
                    break
        if not userAcceptance and ((args is not None and args.answerdefault is not None and str(args.answerdefault).lower() != "y") or (args is not None and args.answerdefault is None)):
            userAcceptance = OutputControls().takeUserInput(f"{colorText.WARN}By using this Software, you agree to\n[+] having read through the Disclaimer {colorText.END}({disclaimerLink}){colorText.WARN}\n[+] and accept Terms Of Service {colorText.END}({tosLink}){colorText.WARN} of PKScreener ? {colorText.END}(Y/N){colorText.GREEN} [Default: {colorText.END}{colorText.FAIL}N{colorText.END}{colorText.GREEN}] :{colorText.END}",defaultInput="N",enableUserInput=True) or "N"
            if str(userAcceptance).lower() != "y":
                OutputControls().printOutput(f"\n{colorText.WARN}You seem to have\n    [+] passed disagreement to the Disclaimer and \n    [+] not accepted Terms Of Service of PKScreener.\n{colorText.END}{colorText.FAIL}[+] You MUST read and agree to the disclaimer and MUST accept the Terms of Service to use PKScreener.{colorText.END}\n\n{colorText.WARN}Exiting now!{colorText.END}")
                sleep(5)
                sys.exit(0)
        try:
            from pkscreener.classes import VERSION
            # Reset logging. If the user indeed passed the --log flag, it will be enabled later anyways
            del os.environ['PKDevTools_Default_Log_Level']
        except: # pragma: no cover
            pass
        configManager.logsEnabled = False
        configManager.tosAccepted = True
        configManager.appVersion = VERSION
        configManager.setConfig(ConfigManager.parser,default=True,showFileCreatedText=False)
        import atexit
        atexit.register(exitGracefully)
        # Set the trigger timestamp
        if args.triggertimestamp is None:
            args.triggertimestamp = int(PKDateUtilities.currentDateTimestamp())
        else:
            args.triggertimestamp = int(args.triggertimestamp)
        # configManager.restartRequestsCache()
        # args.monitor = configManager.defaultMonitorOptions
        if args.monitor is not None:
            from pkscreener.classes.MenuOptions import NA_NON_MARKET_HOURS
            configuredMonitorOptions = configManager.defaultMonitorOptions.split("~") if len(configManager.myMonitorOptions) < 1 else configManager.myMonitorOptions.split("~")
            for monitorOption in NA_NON_MARKET_HOURS:
                if monitorOption in configuredMonitorOptions and not PKDateUtilities.isTradingTime():
                    # These can't be run in non-market hours
                    configuredMonitorOptions.remove(monitorOption)
            MarketMonitor(monitors=args.monitor.split("~") if len(args.monitor) > 5 else configuredMonitorOptions,
                        maxNumResultsPerRow=configManager.maxDashboardWidgetsPerRow,
                        maxNumColsInEachResult=6,
                        maxNumRowsInEachResult=10,
                        maxNumResultRowsInMonitor=configManager.maxNumResultRowsInMonitor,
                        pinnedIntervalWaitSeconds=configManager.pinnedMonitorSleepIntervalSeconds,
                        alertOptions=configManager.soundAlertForMonitorOptions.split("~"))

        if args.log or configManager.logsEnabled:
            setupLogger(shouldLog=True, trace=args.testbuild)
            if not args.prodbuild and args.answerdefault is None:
                try:
                    OutputControls().takeUserInput("Press <Enter> to continue...")
                except EOFError: # pragma: no cover
                    OutputControls().printOutput(f"{colorText.WARN}If you're running from within docker, please run like this:{colorText.END}\n{colorText.FAIL}docker run -it pkjmesra/pkscreener:latest\n{colorText.END}")
                    pass
        else:
            if "PKDevTools_Default_Log_Level" in os.environ.keys():
                del os.environ['PKDevTools_Default_Log_Level']
                # os.environ["PKDevTools_Default_Log_Level"] = str(log.logging.NOTSET)
        if args.simulate:
            os.environ["simulation"] = json.dumps(args.simulate)
        elif "simulation" in os.environ.keys():
            del os.environ['simulation']
        # Import other dependency here because if we import them at the top
        # multiprocessing behaves in unpredictable ways
        from pkscreener.classes import Utility, ConsoleUtility

        configManager.default_logger = default_logger()
        if originalStdOut is None:
            # Clear only if this is the first time it's being called from some
            # loop within workflowtriggers.
            ConsoleUtility.PKConsoleTools.clearScreen(userArgs=args, clearAlways=True)
        warnAboutDependencies()
        if args.prodbuild:
            if args.options and len(args.options.split(":")) > 0:
                indexOption = 0
                doNotDisableGlobalPrint = False
                while indexOption <= 15: # Max integer menu index in level1_X_MenuDict in MenuOptions.py
                    # Menu option 30 is for ATR trailing stops which uses vectorbt
                    # which in turn uses numba which requires print function to be inferred globally
                    # If we try to override print with new_func, it expects this new_func
                    # to be globally available. So to avoid these changes, let's just skip
                    # prodmode for menu option 30
                    doNotDisableGlobalPrint = f":{indexOption}:30:" in args.options
                    if doNotDisableGlobalPrint:
                        break
                    indexOption += 1
                if not doNotDisableGlobalPrint:
                    disableSysOut()
            else:
                disableSysOut()

        if not configManager.checkConfigFile():
            configManager.setConfig(
                ConfigManager.parser, default=True, showFileCreatedText=False
            )
        from pkscreener.classes.PKUserRegistration import PKUserRegistration, ValidationResult
        if args.systemlaunched and not PKUserRegistration.validateToken()[0]:
            result = PKUserRegistration.login()
            if result != ValidationResult.Success:
                OutputControls().printOutput(f"\n[+] {colorText.FAIL}You MUST be a premium/paid user to use this feature!{colorText.END}\n")
                input("Press any key to exit...")
                sys.exit(0)

        if args.systemlaunched and args.options is not None:
            args.systemlaunched = args.options
            
        if args.telegram:
            # Launched by bot for intraday monitor?
            if (PKDateUtilities.isTradingTime() and not PKDateUtilities.isTodayHoliday()[0]) or ("PKDevTools_Default_Log_Level" in os.environ.keys()):
                from PKDevTools.classes import Archiver
                filePath = os.path.join(Archiver.get_user_data_dir(), "monitor_outputs_1.txt")
                if os.path.exists(filePath):
                    default_logger().info("monitor_outputs_1.txt already exists! This means an instance may already be running. Exiting now...")
                    # Since the file exists, it means, there is another instance running
                    return
            else:
                # It should have been launched only during the trading hours
                default_logger().info("--telegram option must be launched ONLY during NSE trading hours. Exiting now...")
                return
        # Check and see if we're running only the telegram bot
        if args.bot:
            from pkscreener import pkscreenerbot
            pkscreenerbot.runpkscreenerbot(availability=args.botavailable)
            return
        updateConfig(args)
        if args.options is not None:
            if str(args.options) == "0":
                # Must be from unit tests to be able to break out of loops via eventing
                args.options = None
            args.options = args.options.replace("::",":")
        
        if args.maxprice:
            configManager.maxLTP = args.maxprice
            configManager.setConfig(ConfigManager.parser, default=True, showFileCreatedText=False)
        if args.minprice:
            configManager.minLTP = args.minprice
            configManager.setConfig(ConfigManager.parser, default=True, showFileCreatedText=False)
        global LoggedIn
        if not LoggedIn and not args.telegram and not args.bot and not args.systemlaunched and not args.testbuild:
            from pkscreener.classes.PKUserRegistration import PKUserRegistration
            if not PKUserRegistration.login():
                sys.exit(0)
            LoggedIn = True
        if args.testbuild and not args.prodbuild:
            OutputControls().printOutput(
                colorText.FAIL
                + "  [+] Started in TestBuild mode!"
                + colorText.END
            )
            runApplication()
            from pkscreener.globals import closeWorkersAndExit
            closeWorkersAndExit()
            exitGracefully()
            sys.exit(0)
        elif args.download:
            OutputControls().printOutput(
                colorText.FAIL
                + "  [+] Download ONLY mode! Stocks will not be screened!"
                + colorText.END
            )
            configManager.restartRequestsCache()
            # if args.intraday is None:
            #     configManager.toggleConfig(candleDuration="1d", clearCache=False)
            runApplication()
            from pkscreener.globals import closeWorkersAndExit
            closeWorkersAndExit()
            exitGracefully()
            sys.exit(0)
        else:
            runApplicationForScreening()
    except KeyboardInterrupt: # pragma: no cover
        from pkscreener.globals import closeWorkersAndExit
        closeWorkersAndExit()
        exitGracefully()
        sys.exit(0)
    except Exception as e: # pragma: no cover
        if "RUNNER" not in os.environ.keys() and ('PKDevTools_Default_Log_Level' in os.environ.keys() and os.environ["PKDevTools_Default_Log_Level"] != str(log.logging.NOTSET)):
                OutputControls().printOutput(
                    "  [+] RuntimeError with 'multiprocessing'.\n  [+] Please contact the Developer, if this does not work!"
                )
                OutputControls().printOutput(e)
                traceback.print_exc()
        pass

def runLoopOnScheduleOrStdApplication(hasCronInterval):
    if hasCronInterval:
        scheduleNextRun()
    else:
        runApplication()

def runApplicationForScreening():
    from pkscreener.globals import closeWorkersAndExit
    try:
        hasCronInterval = args.croninterval is not None and str(args.croninterval).isnumeric()
        shouldBreak = (args.exit and not hasCronInterval)or args.user is not None or args.testbuild
        runLoopOnScheduleOrStdApplication(hasCronInterval)
        while True:
            if shouldBreak:
                break
            runLoopOnScheduleOrStdApplication(hasCronInterval)
        if args.v:
            disableSysOut(disable=False)
            return
        closeWorkersAndExit()
        exitGracefully()
        sys.exit(0)
    except SystemExit: # pragma: no cover
        closeWorkersAndExit()
        exitGracefully()
        sys.exit(0)
    except (RuntimeError, Exception) as e:  # pragma: no cover
        default_logger().debug(e, exc_info=True)
        if args.prodbuild:
            disableSysOut(disable=False)
        OutputControls().printOutput(
            f"{e}\n  [+] An error occurred! Please run with '-l' option to collect the logs.\n  [+] For example, 'pkscreener -l' and then contact the developer!"
        )
        if "RUNNER" in os.environ.keys() or ('PKDevTools_Default_Log_Level' in os.environ.keys() and os.environ["PKDevTools_Default_Log_Level"] != str(log.logging.NOTSET)):
            traceback.print_exc()
        if args.v:
            disableSysOut(disable=False)
            return
        closeWorkersAndExit()
        exitGracefully()
        sys.exit(0)


def scheduleNextRun():
    sleepUntilNextExecution = not PKDateUtilities.isTradingTime()
    while sleepUntilNextExecution:
        OutputControls().printOutput(
            colorText.FAIL
            + (
                "SecondsAfterClosingTime[%d] SecondsBeforeMarketOpen [%d]. Next run at [%s]"
                % (
                    int(PKDateUtilities.secondsAfterCloseTime()),
                    int(PKDateUtilities.secondsBeforeOpenTime()),
                    str(
                        PKDateUtilities.nextRunAtDateTime(
                            bufferSeconds=3600,
                            cronWaitSeconds=int(args.croninterval),
                        )
                    ),
                )
            )
            + colorText.END
        )
        if (PKDateUtilities.secondsAfterCloseTime() >= 3600) and (
            PKDateUtilities.secondsAfterCloseTime() <= (3600 + 1.5 * int(args.croninterval))
        ):
            sleepUntilNextExecution = False
        if (PKDateUtilities.secondsBeforeOpenTime() <= -3600) and (
            PKDateUtilities.secondsBeforeOpenTime() >= (-3600 - 1.5 * int(args.croninterval))
        ):
            sleepUntilNextExecution = False
        sleep(int(args.croninterval))
    global cron_runs
    if cron_runs > 0:
        OutputControls().printOutput(
            colorText.GREEN + f'=> Going to fetch again in {int(args.croninterval)} sec. at {(PKDateUtilities.currentDateTime() + datetime.timedelta(seconds=120)).strftime("%Y-%m-%d %H:%M:%S")} IST...' + colorText.END,
            end="\r",
            flush=True,
        )
        sleep(int(args.croninterval) if not args.testbuild else 3)
    runApplication()
    cron_runs += 1

if __name__ == "__main__":
    if "RUNNER" in os.environ.keys():
        try:
            owner = os.popen('git ls-remote --get-url origin | cut -d/ -f4').read().replace("\n","")
            repo = os.popen('git ls-remote --get-url origin | cut -d/ -f5').read().replace(".git","").replace("\n","")
            if owner.lower() not in ["pkjmesra","pkscreener"]:
                sys.exit(0)
        except:
            pass
    try:
        pkscreenercli()
    except KeyboardInterrupt: # pragma: no cover
        from pkscreener.globals import closeWorkersAndExit
        closeWorkersAndExit()
        exitGracefully()
        sys.exit(0)



"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import pandas as pd
import numpy as np
from PKDevTools.classes.PKDateUtilities import PKDateUtilities
from PKDevTools.classes.ColorText import colorText
from PKDevTools.classes.Singleton import SingletonType, SingletonMixin
from pkscreener.classes.PKScheduledTaskProgress import PKScheduledTaskProgress
from pkscreener.classes.PKTask import PKTask

class PortfolioSecurity:
    def __init__(self, ticker):
        self.name = ticker
        self.ltp = 0
        self.quantity = 0
        self.date = None
        self.growth = 0
    
    @property
    def action(self):
        return (colorText.GREEN + "[Buy]"+ colorText.END) if self.quantity > 0 else ((colorText.FAIL + "[Sell]"+ colorText.END) if self.quantity < 0 else (colorText.WARN + "[Hold]"+ colorText.END))
    
    @property
    def investment(self):
        return self.ltp * self.quantity
    
    @property
    def description(self):
        return {"Date": self.date or PKDateUtilities.currentDateTime().strftime("%Y-%m-%d"),
                "Name": self.name, "LTP": self.ltp, "Quantity": self.quantity, 
                "Action": self.action, "Investment": self.investment, "RunningTotal": 0, 
                "Growth": self.growth, "Profits" : 0}

class Portfolio(PKScheduledTaskProgress):
    def __init__(self, name):
        super(Portfolio, self).__init__()
        self.name = name
        self._initialValue = 0
        self._currentValue = 0
        self.ledger = {}
        self.securities = {}
    
    @property
    def descriptionAsDataframe(self):
        portfolio_df = None
        for date, ledgerEntries in self.ledger.items():
            firstLedgerEntry = ledgerEntries[0]
            if portfolio_df is None:
                portfolio_df = pd.DataFrame(ledgerEntries,columns=firstLedgerEntry.keys())
            else:
                newEntries_df = pd.DataFrame(ledgerEntries,columns=firstLedgerEntry.keys())
                portfolio_df = pd.concat([portfolio_df,newEntries_df], axis=0)
        if portfolio_df is not None:
            portfolio_df["RunningTotal"] = portfolio_df[['Investment']].cumsum()
            portfolio_df["Profits"] = portfolio_df[['Growth']].cumsum()
        return portfolio_df
    
    def updatePortfolioFromXRayDataFrame(self,df:pd.DataFrame, periods:list,task:PKTask=None):
        if task is not None:
            taskId = task.taskId
            if taskId > 0:
                self.tasksDict[taskId] = task
        xray_df = df.copy()
        df_grouped = xray_df.groupby("Stock")
        periodCounter = -1
        task.progress = 0
        for period in periods:
            periodCounter += 1
            if f"LTP{period}" not in xray_df.columns:
                continue
            for stock, df_group in df_grouped:
                df_group["LTP"] = df_group["LTP"].astype(float).fillna(0)
                df_group[f"LTP{period}"] = df_group[f"LTP{period}"].astype(float).fillna(0)
                df_group[f"Growth{period}"] = df_group[f"Growth{period}"].astype(float).fillna(0)
                if df_group.iloc[0][f"LTP{period}"] == 0 or df_group.iloc[0][f"LTP"] == 0:
                    continue
                task.total = len(periods) * len(df_grouped)
                task.progress += 1
                self.updateProgress(task.taskId)
                security = PortfolioSecurity(stock)
                security.ltp = df_group.iloc[0]["LTP"] if not self.hasSecurity(stock) else df_group.iloc[0][f"LTP{period}"]
                previousPeriod = periods[periodCounter-1]
                try:
                    priceRise = round(df_group.iloc[0][f"LTP{period}"] - df_group.iloc[0]["LTP" if periodCounter == 0 else f"LTP{previousPeriod}"],2)
                    growth = df_group.iloc[0][f"Growth{period}"]
                    security.date = df_group.iloc[0]["Date"] if periodCounter == 0 else PKDateUtilities.nextTradingDate(df_group.iloc[0]["Date"], days=period).strftime("%Y-%m-%d")
                    if self.hasSecurity(stock):
                        # This security was already added earlier and exists in the portfolio
                        security.quantity = 1 if priceRise >= 0 else -1
                        if priceRise < 0:
                            security.growth = priceRise * abs(security.quantity)
                            self.removeSecurity(security=security)
                        else:
                            security.quantity = 0 # This is not an actual buy
                            security.growth = priceRise
                            security.ltp = df_group.iloc[0][f"LTP{period}"]
                            self.addSecurity(security=security)
                    else:
                        # This security was never added earlier. The very fact it exists under this
                        # outcome dataframe, we need to take losses and then remove it from portfolio
                        security.quantity = 1
                        security.growth = 0 # First day of trade
                        security.date = df_group.iloc[0]["Date"]
                        security.ltp = df_group.iloc[0]["LTP"]
                        self.addSecurity(security=security)
                        if priceRise < 0:
                            security.date = PKDateUtilities.nextTradingDate(df_group.iloc[0]["Date"], days=period).strftime("%Y-%m-%d")
                            security.ltp = df_group.iloc[0][f"LTP{period}"]
                            security.quantity = -1
                            security.growth = priceRise * abs(security.quantity)
                            self.removeSecurity(security=security)
                except: # pragma: no cover
                    pass
                    continue
        task.progress = task.total
        self.updateProgress(task.taskId)

    @property
    def profit(self):
        sorted_ledger_dates = sorted(self.ledger.items(), key=lambda kv: kv[0])
        bought = 0
        sold = 0
        for date, ledgerEntries in sorted_ledger_dates:
            for securityDict in ledgerEntries:
                if securityDict["Quantity"] > 0 and date == securityDict["Date"]:
                    bought += securityDict["LTP"] * securityDict["Quantity"]
                elif securityDict["Quantity"] < 0 and date == securityDict["Date"]:
                    sold += securityDict["LTP"] * securityDict["Quantity"]
        return round(abs(sold) - bought,2) if (sold != 0 and bought != 0) else 0
        
    @property
    def initialValue(self):
        if self._initialValue != 0:
            return self._initialValue
        sorted_ledger_dates = sorted(self.ledger.items(), key=lambda kv: kv[0])
        initialLedgerEntries = sorted_ledger_dates[1][0]
        initialValue = 0
        for securityDict in initialLedgerEntries:
            initialValue += securityDict["LTP"] * securityDict["Quantity"]
        self._initialValue = initialValue
        return initialValue

    @property
    def currentValue(self):
        if self._currentValue != 0:
            return round(self._currentValue,2)
        sorted_ledger_dates = sorted(self.ledger.items(), key=lambda kv: kv[0])
        currentValue = 0
        for date,ledgerEntries in sorted_ledger_dates:
            for securityDict in ledgerEntries:
                currentValue += securityDict["LTP"] * securityDict["Quantity"]
        self._currentValue = currentValue
        return currentValue
    
    @currentValue.setter
    def currentValue(self, newValue):
        self._currentValue = newValue
        
    def hasSecurity(self, securityName:str):
        return securityName in self.securities.keys()
    
    def addSecurity(self, security:PortfolioSecurity=None):
        self.securities[security.name] = security
        self.currentValue += security.ltp*security.quantity
        self.updateLedger(security=security)

    def removeSecurity(self, security:PortfolioSecurity=None):
        del self.securities[security.name]
        self.currentValue -= security.ltp*security.quantity
        self.updateLedger(security=security)

    def updateLedger(self,security:PortfolioSecurity=None):
        ledgerEntries = self.ledger.get(security.date) or []
        runningLedger = {"ScanType": self.name, "Date": security.date} | security.description
        ledgerEntries.append(runningLedger)
        self.ledger[security.date] = ledgerEntries

    def getDifference(self,x):
        return x.iloc[-1] - x.iloc[0]

    def differenceFromLastNTradingSession(self,df,n=1):
        df['LTP'].rolling(window=n).apply(self.getDifference)


class PortfolioCollection(SingletonMixin,PKScheduledTaskProgress, metaclass=SingletonType):
    def __init__(self):
        super(PortfolioCollection, self).__init__()
        self._portfolios = {}
        self._portfolios_df = None
        self._portfoliosSummary_df = None
        self.ledgerSummaryAsDataframeTaskId = 0
        self.portfoliosAsDataframeTaskId = 0

    def getLedgerSummaryAsDataframe(self,*args):
        task = args[0]
        taskId = task.taskId
        if taskId > 0:
            self.tasksDict[taskId] = task
            self.ledgerSummaryAsDataframeTaskId = taskId
        return self.ledgerSummaryAsDataframe
    
    def getPortfoliosAsDataframe(self,*args):
        task = args[0]
        taskId = task.taskId
        if taskId > 0:
            self.tasksDict[taskId] = task
            self.portfoliosAsDataframeTaskId = taskId
        return self.portfoliosAsDataframe
    
    @property
    def ledgerSummaryAsDataframe(self):
        task = None
        if self._portfoliosSummary_df is not None:
            if self.ledgerSummaryAsDataframeTaskId > 0:
                task = self.tasksDict.get(self.ledgerSummaryAsDataframeTaskId)
                task.total = len(self._portfolios.keys())
                task.progress = task.total
                self.updateProgress(self.ledgerSummaryAsDataframeTaskId)
                task.result = self._portfoliosSummary_df
                task.resultsDict[task.taskId] = self._portfoliosSummary_df
            return self._portfoliosSummary_df
        portfolios_df = None
        if len(self._portfolios) > 0:
            for _, portfolio in self._portfolios.items():
                portfolio_df = portfolio.descriptionAsDataframe
                if portfolio_df is None:
                    continue
                if portfolios_df is None:
                    portfolios_df = portfolio_df.tail(5).copy()
                else:
                    portfolios_df = pd.concat([portfolios_df,portfolio_df.tail(5)], axis=0)
                if task is not None:
                    task.progress = len(portfolios_df)
                self.updateProgress(self.ledgerSummaryAsDataframeTaskId)
        self._portfoliosSummary_df = portfolios_df
        if task is not None:
            # Mark the progress finished
            task.progress = task.total
            self.updateProgress(self.ledgerSummaryAsDataframeTaskId)
            task.result = portfolios_df
            task.resultsDict[task.taskId] = self.portfolios_df
        return portfolios_df
        
    @property
    def portfoliosAsDataframe(self):
        task = None
        if self._portfolios_df is not None:
            if self.portfoliosAsDataframeTaskId > 0:
                task = self.tasksDict.get(self.portfoliosAsDataframeTaskId)
                task.total = len(self._portfolios.keys())
                task.progress = task.total
                self.updateProgress(self.ledgerSummaryAsDataframeTaskId)
                task.result = self._portfolios_df
                task.resultsDict[task.taskId] = self._portfolios_df
            return self._portfolios_df
        portfolios_df = None
        if len(self._portfolios) > 0:
            for _, portfolio in self._portfolios.items():
                portfolio_df = portfolio.descriptionAsDataframe
                if portfolio_df is None:
                    continue
                if portfolios_df is None:
                    portfolios_df = portfolio_df.copy()
                else:
                    portfolios_df = pd.concat([portfolios_df,portfolio_df], axis=0)
                if task is not None:
                    task.progress = len(portfolios_df)
                self.updateProgress(self.ledgerSummaryAsDataframeTaskId)
        self._portfolios_df = portfolios_df
        if task is not None:
            # Mark the progress finished
            task.progress = task.total
            self.updateProgress(self.ledgerSummaryAsDataframeTaskId)
            task.result = portfolios_df
            task.resultsDict[task.taskId] = self.portfolios_df
        return portfolios_df

    def addPortfolio(self,portfolio:Portfolio):
        self._portfolios[portfolio.name] = portfolio
